# -*- coding: utf-8 -*-
import logging
import utool as ut
import numpy as np
import vtool as vt
import functools
from wbia.algo.hots import hstypes
from wbia.algo.hots import _pipeline_helpers as plh
from six.moves import zip, range, map  # NOQA

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')

"""
FIXME: qfx2_ no longer applies due to fgw_thresh. Need to change names in this file

TODO: replace testdata_pre_weight_neighbors with

        >>> qreq_, args = plh.testdata_pre('weight_neighbors', defaultdb='testdb1',
        >>>                                a=['default:qindex=0:1,dindex=0:5,hackerrors=False'],
        >>>                                p=['default:codename=vsmany,bar_l2_on=True,fg_on=False'], verbose=True)

"""


NN_WEIGHT_FUNC_DICT = {}
MISC_WEIGHT_FUNC_DICT = {}
EPS = 1e-8


def _register_nn_normalized_weight_func(func):
    r"""
    Decorator for weighting functions

    Registers a nearest neighbor normalized weighting
    Used for LNBNN
    """
    global NN_WEIGHT_FUNC_DICT
    filtkey = ut.get_funcname(func).replace('_fn', '').lower()
    if ut.VERYVERBOSE:
        logger.info('[nn_weights] registering norm func: %r' % (filtkey,))
    filtfunc = functools.partial(nn_normalized_weight, func)
    NN_WEIGHT_FUNC_DICT[filtkey] = filtfunc
    return func


def _register_nn_simple_weight_func(func):
    """
    Used for things that dont require a normalizer like const
    """
    filtkey = ut.get_funcname(func).replace('_match_weighter', '').lower()
    if ut.VERYVERBOSE:
        logger.info('[nn_weights] registering simple func: %r' % (filtkey,))
    NN_WEIGHT_FUNC_DICT[filtkey] = func
    return func


def _register_misc_weight_func(func):
    filtkey = ut.get_funcname(func).replace('_match_weighter', '').lower()
    if ut.VERYVERBOSE:
        logger.info('[nn_weights] registering simple func: %r' % (filtkey,))
    MISC_WEIGHT_FUNC_DICT[filtkey] = func
    return func


@_register_nn_simple_weight_func
def const_match_weighter(nns_list, nnvalid0_list, qreq_):
    r"""
    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> #tup = plh.testdata_pre_weight_neighbors('PZ_MTEST')
        >>> qreq_, args = plh.testdata_pre('weight_neighbors', defaultdb='PZ_MTEST')
        >>> nns_list, nnvalid0_list = args
        >>> ibs, qreq_, nns_list, nnvalid0_list = tup
        >>> constvote_weight_list = const_match_weighter(nns_list, nnvalid0_list, qreq_)
        >>> result = ('constvote_weight_list = %s' % (str(constvote_weight_list),))
        >>> print(result)
    """
    constvote_weight_list = []
    # K = qreq_.qparams.K  # Dont use K because K is dynamic per query
    # Subtract Knorm from size instead
    Knorm = qreq_.qparams.Knorm

    for nns in nns_list:
        (neighb_idx, neighb_dist) = nns
        neighb_constvote = np.ones(
            (len(neighb_idx), len(neighb_idx.T) - Knorm), dtype=np.float
        )
        constvote_weight_list.append(neighb_constvote)
    return constvote_weight_list


@_register_nn_simple_weight_func
def fg_match_weighter(nns_list, nnvalid0_list, qreq_):
    r"""
    foreground feature match weighting

    CommandLine:
        python -m wbia.algo.hots.nn_weights --exec-fg_match_weighter

    Example:
        >>> # xdoctest: +REQUIRES(module:wbia_cnn)
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> #tup = plh.testdata_pre_weight_neighbors('PZ_MTEST')
        >>> #ibs, qreq_, nns_list, nnvalid0_list = tup
        >>> qreq_, args = plh.testdata_pre('weight_neighbors', defaultdb='PZ_MTEST')
        >>> nns_list, nnvalid0_list = args
        >>> print(ut.repr2(qreq_.qparams.__dict__, sorted_=True))
        >>> assert qreq_.qparams.fg_on == True, 'bug setting custom params fg_on'
        >>> fgvotes_list = fg_match_weighter(nns_list, nnvalid0_list, qreq_)
        >>> print('fgvotes_list = %r' % (fgvotes_list,))
    """
    Knorm = qreq_.qparams.Knorm
    config2_ = qreq_.get_internal_query_config2()
    # Database feature index to chip index
    fgvotes_list = []
    for nn in nns_list:
        # database forground weights
        neighb_dfgws = qreq_.indexer.get_nn_fgws(nn.neighb_idxs.T[0:-Knorm].T)
        # query forground weights
        qfx2_qfgw = qreq_.ibs.get_annot_fgweights(
            [nn.qaid], ensure=False, config2_=config2_
        )[0]
        qfgws = qfx2_qfgw.take(nn.qfx_list, axis=0)
        # feature match forground weight is geometric mean
        neighb_fgvote_weight = np.sqrt(qfgws[:, None] * neighb_dfgws)
        fgvotes_list.append(neighb_fgvote_weight)
    return fgvotes_list


def nn_normalized_weight(normweight_fn, nns_list, nnvalid0_list, qreq_):
    r"""
    Generic function to weight nearest neighbors

    ratio, lnbnn, and other nearest neighbor based functions use this

    Args:
        normweight_fn (func): chosen weight function e.g. lnbnn
        nns_list (dict): query descriptor nearest neighbors and distances.
        nnvalid0_list (list): list of neighbors preflagged as valid
        qreq_ (QueryRequest): hyper-parameters

    Returns:
        list: weights_list

    CommandLine:
        python -m wbia.algo.hots.nn_weights nn_normalized_weight --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> #tup = plh.testdata_pre_weight_neighbors('PZ_MTEST')
        >>> #ibs, qreq_, nns_list, nnvalid0_list = tup
        >>> qreq_, args = plh.testdata_pre('weight_neighbors',
        >>>                                defaultdb='PZ_MTEST')
        >>> nns_list, nnvalid0_list = args
        >>> normweight_fn = lnbnn_fn
        >>> weights_list1, normk_list1 = nn_normalized_weight(
        >>>     normweight_fn, nns_list, nnvalid0_list, qreq_)
        >>> weights1 = weights_list1[0]
        >>> nn_normonly_weight = NN_WEIGHT_FUNC_DICT['lnbnn']
        >>> weights_list2, normk_list2 = nn_normonly_weight(nns_list, nnvalid0_list, qreq_)
        >>> weights2 = weights_list2[0]
        >>> assert np.all(weights1 == weights2)
        >>> ut.assert_inbounds(weights1.sum(), 100, 510)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> #tup = plh.testdata_pre_weight_neighbors('PZ_MTEST')
        >>> qreq_, args = plh.testdata_pre('weight_neighbors',
        >>>                                defaultdb='PZ_MTEST')
        >>> nns_list, nnvalid0_list = args
        >>> normweight_fn = ratio_fn
        >>> weights_list1, normk_list1 = nn_normalized_weight(normweight_fn, nns_list, nnvalid0_list, qreq_)
        >>> weights1 = weights_list1[0]
        >>> nn_normonly_weight = NN_WEIGHT_FUNC_DICT['ratio']
        >>> weights_list2, normk_list2 = nn_normonly_weight(nns_list, nnvalid0_list, qreq_)
        >>> weights2 = weights_list2[0]
        >>> assert np.all(weights1 == weights2)
        >>> ut.assert_inbounds(weights1.sum(), 1500, 4500)
    """
    Knorm = qreq_.qparams.Knorm
    normalizer_rule = qreq_.qparams.normalizer_rule
    # Database feature index to chip index
    qaid_list = qreq_.get_internal_qaids()
    normk_list = [
        get_normk(qreq_, qaid, neighb_idx, Knorm, normalizer_rule)
        for qaid, (neighb_idx, neighb_dist) in zip(qaid_list, nns_list)
    ]
    weight_list = [
        apply_normweight(normweight_fn, neighb_normk, neighb_idx, neighb_dist, Knorm)
        for neighb_normk, (neighb_idx, neighb_dist) in zip(normk_list, nns_list)
    ]
    return weight_list, normk_list


def get_normk(qreq_, qaid, neighb_idx, Knorm, normalizer_rule):
    """
    Get positions of the LNBNN/ratio tests normalizers

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> cfgdict = {'K':10, 'Knorm': 10, 'normalizer_rule': 'name',
        >>>            'dim_size': 450, 'resize_dim': 'area'}
        >>> #tup = plh.testdata_pre_weight_neighbors(cfgdict=cfgdict)
        >>> qreq_, args = plh.testdata_pre('weight_neighbors', defaultdb='testdb1',
        >>>                                p=['default:K=10,Knorm=10,normalizer_rule=name,dim_size=450,resize_dim=area'])
        >>> nns_list, nnvalid0_list = args
        >>> (neighb_idx, neighb_dist) = nns_list[0]
        >>> qaid = qreq_.qaids[0]
        >>> K = qreq_.qparams.K
        >>> Knorm = qreq_.qparams.Knorm
        >>> neighb_normk1 = get_normk(qreq_, qaid, neighb_idx, Knorm, 'last')
        >>> neighb_normk2 = get_normk(qreq_, qaid, neighb_idx, Knorm, 'name')
        >>> assert np.all(neighb_normk1 == Knorm + K)
        >>> assert np.all(neighb_normk2 <= Knorm + K) and np.all(neighb_normk2 > K)
    """
    K = len(neighb_idx.T) - Knorm
    assert K > 0, 'K=%r cannot be 0' % (K,)
    # neighb_nndist = neighb_dist.T[0:K].T
    if normalizer_rule == 'last':
        neighb_normk = np.zeros(len(neighb_idx), hstypes.FK_DTYPE) + (K + Knorm - 1)
    elif normalizer_rule == 'name':
        neighb_normk = get_name_normalizers(qaid, qreq_, Knorm, neighb_idx)
    elif normalizer_rule == 'external':
        pass
    else:
        raise NotImplementedError('[nn_weights] no normalizer_rule=%r' % normalizer_rule)
    return neighb_normk


def apply_normweight(normweight_fn, neighb_normk, neighb_idx, neighb_dist, Knorm):
    r"""
    helper applies the normalized weight function to one query annotation

    Args:
        normweight_fn (func):  chosen weight function e.g. lnbnn
        qaid (int):  query annotation id
        neighb_idx (ndarray[int32_t, ndims=2]):  mapping from query feature
            index to db neighbor index
        neighb_dist (ndarray):  mapping from query feature index to dist
        Knorm (int):
        qreq_ (QueryRequest):  query request object with hyper-parameters

    Returns:
        ndarray: neighb_normweight

    CommandLine:
        python -m wbia.algo.hots.nn_weights --test-apply_normweight

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> from wbia.algo.hots import nn_weights
        >>> #cfgdict = {'K':10, 'Knorm': 10, 'normalizer_rule': 'name',
        >>> #           'dim_size': 450, 'resize_dim': 'area'}
        >>> #tup = plh.testdata_pre_weight_neighbors(cfgdict=cfgdict)
        >>> qreq_, args = plh.testdata_pre('weight_neighbors', defaultdb='testdb1',
        >>>                                p=['default:K=10,Knorm=10,normalizer_rule=name,dim_size=450,resize_dim=area'])
        >>> nns_list, nnvalid0_list = args
        >>> qaid = qreq_.qaids[0]
        >>> Knorm = qreq_.qparams.Knorm
        >>> normweight_fn = lnbnn_fn
        >>> normalizer_rule  = qreq_.qparams.normalizer_rule
        >>> (neighb_idx, neighb_dist) = nns_list[0]
        >>> neighb_normk = get_normk(qreq_, qaid, neighb_idx, Knorm, normalizer_rule)
        >>> neighb_normweight = nn_weights.apply_normweight(
        >>>   normweight_fn, neighb_normk, neighb_idx, neighb_dist, Knorm)
        >>> ut.assert_inbounds(neighb_normweight.sum(), 600, 950)
    """
    K = len(neighb_idx.T) - Knorm
    # neighb_normdist = np.array(
    #     [dists[normk] for (dists, normk) in zip(neighb_dist, neighb_normk)])
    neighb_normdist = vt.take_col_per_row(neighb_dist, neighb_normk)
    neighb_normdist.shape = (len(neighb_idx), 1)
    neighb_nndist = neighb_dist.T[0:K].T
    vdist = neighb_nndist  # voting distance
    ndist = neighb_normdist  # normalizer distance
    neighb_normweight = normweight_fn(vdist, ndist)
    return neighb_normweight


def get_name_normalizers(qaid, qreq_, Knorm, neighb_idx):
    r"""
    helper normalizers for 'name' normalizer_rule

    Args:
        qaid (int): query annotation id
        qreq_ (wbia.QueryRequest): hyper-parameters
        Knorm (int):
        neighb_idx (ndarray):

    Returns:
        ndarray : neighb_normk

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> from wbia.algo.hots import nn_weights
        >>> #cfgdict = {'K':10, 'Knorm': 10, 'normalizer_rule': 'name'}
        >>> #tup = plh.testdata_pre_weight_neighbors(cfgdict=cfgdict)
        >>> qreq_, args = plh.testdata_pre('weight_neighbors', defaultdb='testdb1',
        >>>                                p=['default:K=10,Knorm=10,normalizer_rule=name'])
        >>> nns_list, nnvalid0_list = args
        >>> Knorm = qreq_.qparams.Knorm
        >>> (neighb_idx, neighb_dist) = nns_list[0]
        >>> qaid = qreq_.qaids[0]
        >>> neighb_normk = get_name_normalizers(qaid, qreq_, Knorm, neighb_idx)
    """
    assert Knorm == qreq_.qparams.Knorm, 'inconsistency in qparams'
    # Get the top names you do not want your normalizer to be from
    # qnid = qreq_.internal_qannots.loc([qaid]).nids[0]
    qnid = qreq_.get_qreq_annot_nids(qaid)
    K = len(neighb_idx.T) - Knorm
    assert K > 0, 'K cannot be 0'
    # Get the 0th - Kth matching neighbors
    neighb_topidx = neighb_idx.T[0:K].T
    # Get tke Kth - KNth normalizing neighbors
    neighb_normidx = neighb_idx.T[-Knorm:].T
    # Apply temporary uniquish name
    neighb_topaid = qreq_.indexer.get_nn_aids(neighb_topidx)
    neighb_normaid = qreq_.indexer.get_nn_aids(neighb_normidx)
    neighb_topnid = qreq_.get_qreq_annot_nids(neighb_topaid)
    neighb_normnid = qreq_.get_qreq_annot_nids(neighb_normaid)
    # Inspect the potential normalizers
    neighb_selnorm = mark_name_valid_normalizers(qnid, neighb_topnid, neighb_normnid)
    neighb_normk = neighb_selnorm + (K + Knorm)  # convert form negative to pos indexes
    return neighb_normk


def mark_name_valid_normalizers(qnid, neighb_topnid, neighb_normnid):
    r"""
    Helper func that allows matches only to the first result for a name

    Each query feature finds its K matches and Kn normalizing matches. These
    are the candidates from which it can choose a set of matches and a single
    normalizer.

    A normalizer is marked as invalid if it belongs to a name that was also in
    its feature's candidate matching set.

    Args:
        neighb_topnid (ndarray): marks the names a feature matches
        neighb_normnid (ndarray): marks the names of the feature normalizers
        qnid (int): query name id

    Returns:
        neighb_selnorm - index of the selected normalizer for each query feature

    CommandLine:
        python -m wbia.algo.hots.nn_weights --exec-mark_name_valid_normalizers

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> qnid = 1
        >>> neighb_topnid = np.array([[1, 1, 1, 1, 1],
        ...                         [1, 2, 1, 1, 1],
        ...                         [1, 2, 2, 3, 1],
        ...                         [5, 8, 9, 8, 8],
        ...                         [5, 8, 9, 8, 8],
        ...                         [6, 6, 9, 6, 8],
        ...                         [5, 8, 6, 6, 6],
        ...                         [1, 2, 8, 6, 6]], dtype=np.int32)
        >>> neighb_normnid = np.array([[ 1, 1, 1],
        ...                          [ 2, 3, 1],
        ...                          [ 2, 3, 1],
        ...                          [ 6, 6, 6],
        ...                          [ 6, 6, 8],
        ...                          [ 2, 6, 6],
        ...                          [ 6, 6, 1],
        ...                          [ 4, 4, 9]], dtype=np.int32)
        >>> neighb_selnorm = mark_name_valid_normalizers(qnid, neighb_topnid, neighb_normnid)
        >>> K = len(neighb_topnid.T)
        >>> Knorm = len(neighb_normnid.T)
        >>> neighb_normk_ = neighb_selnorm + (Knorm)  # convert form negative to pos indexes
        >>> result = str(neighb_normk_)
        >>> print(result)
        [2 1 2 0 0 0 2 0]

    Ignore:
        logger.info(ut.doctest_repr(neighb_normnid, 'neighb_normnid', verbose=False))
        logger.info(ut.doctest_repr(neighb_topnid, 'neighb_topnid', verbose=False))
    """
    # TODO?: warn if any([np.any(flags) for flags in neighb_invalid]), (
    #    'Normalizers are potential matches. Increase Knorm')
    neighb_valid = np.logical_and.reduce(
        [col1[:, None] != neighb_normnid for col1 in neighb_topnid.T]
    )
    # Mark self as invalid, if given that information
    neighb_valid = np.logical_and(neighb_normnid != qnid, neighb_valid)
    # For each query feature find its best normalizer (using negative indices)
    Knorm = neighb_normnid.shape[1]
    neighb_validxs = [np.nonzero(normrow)[0] for normrow in neighb_valid]
    neighb_selnorm = np.array(
        [validxs[0] - Knorm if len(validxs) != 0 else -1 for validxs in neighb_validxs],
        hstypes.FK_DTYPE,
    )
    return neighb_selnorm


@_register_nn_normalized_weight_func
def lnbnn_fn(vdist, ndist):
    r"""
    Locale Naive Bayes Nearest Neighbor weighting

    References:
        http://www.cs.ubc.ca/~lowe/papers/12mccannCVPR.pdf
        http://www.cs.ubc.ca/~sanchom/local-naive-bayes-nearest-neighbor

    Sympy:
        >>> import sympy
        >>> #https://github.com/sympy/sympy/pull/10247
        >>> from sympy import log
        >>> from sympy.stats import P, E, variance, Die, Normal, FiniteRV
        >>> C, Cbar = sympy.symbols('C Cbar')
        >>> d_i = Die(sympy.symbols('di'), 6)
        >>> log(P(di, C) / P(di, Cbar))
        >>> #
        >>> PdiC, PdiCbar = sympy.symbols('PdiC, PdiCbar')
        >>> oddsC = log(PdiC / PdiCbar)
        >>> sympy.simplify(oddsC)
        >>> import vtool as vt
        >>> vt.check_expr_eq(oddsC, log(PdiC) - log(PdiCbar))


    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> out = lnbnn_fn(vdist, ndist)
        >>> result = ut.hz_str('lnbnn  = ', ut.repr2(out, precision=2))
        >>> print(result)
        lnbnn  = np.array([[0.62, 0.22, 0.03],
                           [0.35, 0.22, 0.01],
                           [0.87, 0.58, 0.27],
                           [0.67, 0.42, 0.25],
                           [0.59, 0.3 , 0.27]])
    """
    return ndist - vdist


@_register_nn_normalized_weight_func
def ratio_fn(vdist, ndist):
    r"""
    Args:
        vdist (ndarray): voting array
        ndist (ndarray): normalizing array

    Returns:
        ndarray: out

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> out = ratio_fn(vdist, ndist)
        >>> result = ut.hz_str('ratio = ', ut.repr2(out, precision=2))
        >>> print(result)
        ratio = np.array([[0.  , 0.65, 0.95],
                          [0.33, 0.58, 0.98],
                          [0.13, 0.42, 0.73],
                          [0.15, 0.47, 0.68],
                          [0.23, 0.61, 0.65]])
    """
    return np.divide(vdist, ndist)


@_register_nn_normalized_weight_func
def bar_l2_fn(vdist, ndist):
    r"""
    The feature weight is (1 - the euclidian distance
    between the features). The normalizers are unused.

    (not really a normaalized function)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> out = bar_l2_fn(vdist, ndist)
        >>> result = ut.hz_str('barl2  = ', ut.repr2(out, precision=2))
        >>> print(result)
        barl2  = np.array([[1.  , 0.6 , 0.41],
                           [0.83, 0.7 , 0.49],
                           [0.87, 0.58, 0.27],
                           [0.88, 0.63, 0.46],
                           [0.82, 0.53, 0.5 ]])
    """
    return 1.0 - vdist


@_register_nn_normalized_weight_func
def loglnbnn_fn(vdist, ndist):
    r"""
    Ignore:
        import vtool as vt
        vt.check_expr_eq('log(d) - log(n)', 'log(d / n)')   # True
        vt.check_expr_eq('log(d) / log(n)', 'log(d - n)')

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> out = loglnbnn_fn(vdist, ndist)
        >>> result = ut.hz_str('loglnbnn  = ', ut.repr2(out, precision=2))
        >>> print(result)
        loglnbnn  = np.array([[0.48, 0.2 , 0.03],
                              [0.3 , 0.2 , 0.01],
                              [0.63, 0.46, 0.24],
                              [0.51, 0.35, 0.22],
                              [0.46, 0.26, 0.24]])
    """
    return np.log(ndist - vdist + 1.0)


@_register_nn_normalized_weight_func
def logratio_fn(vdist, ndist):
    r"""
    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> out = normonly_fn(vdist, ndist)
        >>> result = ut.repr2(out)
        >>> print(result)
        np.array([[0.62, 0.62, 0.62],
                  [0.52, 0.52, 0.52],
                  [1.  , 1.  , 1.  ],
                  [0.79, 0.79, 0.79],
                  [0.77, 0.77, 0.77]])
    """
    return np.log(np.divide(ndist, vdist + EPS) + 1.0)


@_register_nn_normalized_weight_func
def normonly_fn(vdist, ndist):
    r"""
    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> out = normonly_fn(vdist, ndist)
        >>> result = ut.repr2(out)
        >>> print(result)
        np.array([[0.62, 0.62, 0.62],
                  [0.52, 0.52, 0.52],
                  [1.  , 1.  , 1.  ],
                  [0.79, 0.79, 0.79],
                  [0.77, 0.77, 0.77]])
    """
    return np.tile(ndist[:, 0:1], (1, vdist.shape[1]))
    # return ndist[None, 0:1]


def testdata_vn_dists(nfeats=5, K=3):
    r"""
    Test voting and normalizing distances

    Returns:
        tuple : (vdist, ndist) - test voting distances and normalizer distances

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> vdist, ndist = testdata_vn_dists()
        >>> result = (ut.hz_str('vdist = ', ut.repr2(vdist))) + '\n'
        >>> print(result + (ut.hz_str('ndist = ', ut.repr2(ndist))))
        vdist = np.array([[0.  , 0.4 , 0.59],
                          [0.17, 0.3 , 0.51],
                          [0.13, 0.42, 0.73],
                          [0.12, 0.37, 0.54],
                          [0.18, 0.47, 0.5 ]])
        ndist = np.array([[0.62],
                          [0.52],
                          [1.  ],
                          [0.79],
                          [0.77]])
    """

    def make_precise(dist):
        prec = 100
        dist = (prec * dist).astype(np.uint8) / prec
        dist = dist.astype(hstypes.FS_DTYPE)
        return dist

    rng = np.random.RandomState(0)
    vdist = rng.rand(nfeats, K)
    ndist = rng.rand(nfeats, 1)
    # Ensure distance increases
    vdist = vdist.cumsum(axis=1)
    ndist = (ndist.T + vdist.max(axis=1)).T
    Z = ndist.max()
    vdist = make_precise(vdist / Z)
    ndist = make_precise(ndist / Z)
    vdist[0][0] = 0
    return vdist, ndist


# @_register_nn_normalized_weight_func
# def dist_fn(vdist, ndist):
#    """ the euclidian distance between the features """
#    return vdist


# @_register_nn_simple_weight_func
def gravity_match_weighter(nns_list, nnvalid0_list, qreq_):
    raise NotImplementedError('have not finished gv weighting')
    # qfx2_nnkpts = qreq_.indexer.get_nn_kpts(qfx2_nnidx)
    # qfx2_nnori = ktool.get_oris(qfx2_nnkpts)
    # qfx2_kpts  = qreq_.ibs.get_annot_kpts(qaid, config2_=qreq_.get_internal_query_config2())  # FIXME: Highly inefficient
    # qfx2_oris  = ktool.get_oris(qfx2_kpts)
    # # Get the orientation distance
    # qfx2_oridist = vt.rowwise_oridist(qfx2_nnori, qfx2_oris)
    # # Normalize into a weight (close orientations are 1, far are 0)
    # qfx2_gvweight = (TAU - qfx2_oridist) / TAU
    # # Apply gravity vector weight to the score
    # qfx2_score *= qfx2_gvweight


def all_normalized_weights_test():
    r"""
    CommandLine:
        python -m wbia.algo.hots.nn_weights --exec-all_normalized_weights_test

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.hots.nn_weights import *  # NOQA
        >>> all_normalized_weights_test()
    """
    from wbia.algo.hots import nn_weights
    import six

    # ibs, qreq_, nns_list, nnvalid0_list = plh.testdata_pre_weight_neighbors()

    qreq_, args = plh.testdata_pre(
        'weight_neighbors',
        defaultdb='testdb1',
        a=['default:qindex=0:1,dindex=0:5,hackerrors=False'],
        p=['default:codename=vsmany,bar_l2_on=True,fg_on=False'],
        verbose=True,
    )
    nns_list = args.nns_list
    nnvalid0_list = args.nnvalid0_list
    qaid = qreq_.qaids[0]

    def tst_weight_fn(nn_weight, nns_list, qreq_, qaid):
        normweight_fn = nn_weights.__dict__[nn_weight + '_fn']
        weight_list1, nomx_list1 = nn_weights.nn_normalized_weight(
            normweight_fn, nns_list, nnvalid0_list, qreq_
        )
        weights1 = weight_list1[0]
        # ---
        # test NN_WEIGHT_FUNC_DICT
        # ---
        nn_normonly_weight = nn_weights.NN_WEIGHT_FUNC_DICT[nn_weight]
        weight_list2, nomx_list2 = nn_normonly_weight(nns_list, nnvalid0_list, qreq_)
        weights2 = weight_list2[0]
        assert np.all(weights1 == weights2)
        logger.info(nn_weight + ' passed')

    for nn_weight in six.iterkeys(nn_weights.NN_WEIGHT_FUNC_DICT):
        normweight_key = nn_weight + '_fn'
        if normweight_key not in nn_weights.__dict__:
            continue
        tst_weight_fn(nn_weight, nns_list, qreq_, qaid)



# -*- coding: utf-8 -*-
import logging
import six
import numpy as np
import utool as ut
import ubelt as ub
import pandas as pd
import itertools as it
import wbia.constants as const
from wbia.algo.graph.state import POSTV, NEGTV, INCMP, NULL
from wbia.algo.graph.refresh import RefreshCriteria

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


PRINCETON_KAIA_EDGE_LIST = None


class InfrLoops(object):
    """
    Algorithm control flow loops
    """

    def main_gen(infr, max_loops=None, use_refresh=True):
        """
        The main outer loop.

        This function is designed as an iterator that will execute the graph
        algorithm main loop as automatically as possible, but if user input is
        needed, it will pause and yield the decision it needs help with. Once
        feedback is given for this item, you can continue the main loop by
        calling next. StopIteration is raised once the algorithm is complete.

        Args:
            max_loops(int): maximum number of times to run the outer loop,
                i.e. ranking is run at most this many times.
            use_refresh(bool): allow the refresh criterion to stop the algo

        Notes:
            Different phases of the main loop are implemented as subiterators

        CommandLine:
            python -m wbia.algo.graph.mixin_loops main_gen

        Doctest:
            >>> from wbia.algo.graph.mixin_loops import *
            >>> from wbia.algo.graph.mixin_simulation import UserOracle
            >>> import wbia
            >>> infr = wbia.AnnotInference('testdb1', aids='all',
            >>>                             autoinit='staging', verbose=4)
            >>> infr.params['manual.n_peek'] = 10
            >>> infr.params['ranking.ntop'] = 1
            >>> infr.oracle = UserOracle(.99, rng=0)
            >>> infr.simulation_mode = False
            >>> infr.reset()
            >>> #infr.load_published()
            >>> gen = infr.main_gen()
            >>> while True:
            >>>     try:
            >>>         reviews = next(gen)
            >>>         edge, priority, data = reviews[0]
            >>>         feedback = infr.request_oracle_review(edge)
            >>>         infr.add_feedback(edge, **feedback)
            >>>     except StopIteration:
            >>>         break
        """
        infr.print('Starting main loop', 1)
        infr.print('infr.params = {}'.format(ut.repr3(infr.params)))
        if max_loops is None:
            max_loops = infr.params['algo.max_outer_loops']
            if max_loops is None:
                max_loops = np.inf

        if infr.test_mode:
            logger.info('------------------ {} -------------------'.format(infr.name))

        # Initialize a refresh criteria
        infr.init_refresh()

        infr.phase = 0
        # Phase 0.1: Ensure the user sees something immediately
        if infr.params['algo.quickstart']:
            infr.loop_phase = 'quickstart_init'
            # quick startup. Yield a bunch of random edges
            num = infr.params['manual.n_peek']
            user_request = []
            for edge in ut.random_combinations(infr.aids, 2, num=num):
                user_request += [infr._make_review_tuple(edge, None)]
                yield user_request

        if infr.params['algo.hardcase']:
            infr.loop_phase = 'hardcase_init'
            # Check previously labeled edges that where the groundtruth and the
            # verifier disagree.
            for _ in infr.hardcase_review_gen():
                yield _

        if infr.params['inference.enabled']:
            infr.loop_phase = 'incon_recover_init'
            # First, fix any inconsistencies
            for _ in infr.incon_recovery_gen():
                yield _

        # Phase 0.2: Ensure positive redundancy (this is generally quick)
        # so the user starts seeing real work after one random review is made
        # unless the graph is already positive redundant.
        if infr.params['redun.enabled'] and infr.params['redun.enforce_pos']:
            infr.loop_phase = 'pos_redun_init'
            # Fix positive redundancy of anything within the loop
            for _ in infr.pos_redun_gen():
                yield _

        infr.phase = 1
        if infr.params['ranking.enabled']:
            for count in it.count(0):

                infr.print('Outer loop iter %d ' % (count,))

                # Phase 1: Try to merge PCCs by searching for LNBNN candidates
                infr.loop_phase = 'ranking_{}'.format(count)
                for _ in infr.ranked_list_gen(use_refresh):
                    yield _

                terminate = infr.refresh.num_meaningful == 0
                if terminate:
                    infr.print('Triggered break criteria', 1, color='red')

                # Phase 2: Ensure positive redundancy.
                infr.phase = 2
                infr.loop_phase = 'posredun_{}'.format(count)
                if all(ut.take(infr.params, ['redun.enabled', 'redun.enforce_pos'])):
                    # Fix positive redundancy of anything within the loop
                    for _ in infr.pos_redun_gen():
                        yield _

                logger.info('prob_any_remain = %r' % (infr.refresh.prob_any_remain(),))
                logger.info(
                    'infr.refresh.num_meaningful = {!r}'.format(
                        infr.refresh.num_meaningful
                    )
                )

                if (count + 1) >= max_loops:
                    infr.print('early stop', 1, color='red')
                    break

                if terminate:
                    infr.print('break triggered')
                    break

        infr.phase = 3
        # Phase 0.3: Ensure positive redundancy (this is generally quick)
        if all(ut.take(infr.params, ['redun.enabled', 'redun.enforce_neg'])):
            # Phase 3: Try to automatically acheive negative redundancy without
            # asking the user to do anything but resolve inconsistency.
            infr.print('Entering phase 3', 1, color='red')
            infr.loop_phase = 'negredun'
            for _ in infr.neg_redun_gen():
                yield _

        infr.phase = 4
        infr.print('Terminate', 1, color='red')
        infr.print('Exiting main loop')

        if infr.params['inference.enabled']:
            infr.assert_consistency_invariant()

    def hardcase_review_gen(infr):
        """
        Subiterator for hardcase review

        Re-review non-confident edges that vsone did not classify correctly
        """
        infr.print('==============================', color='white')
        infr.print('--- HARDCASE PRIORITY LOOP ---', color='white')

        verifiers = infr.learn_evaluation_verifiers()
        verif = verifiers['match_state']

        edges_ = list(infr.edges())
        real_ = list(infr.edge_decision_from(edges_))
        flags_ = [r in {POSTV, NEGTV, INCMP} for r in real_]
        real = ut.compress(real_, flags_)
        edges = ut.compress(edges_, flags_)

        hardness = 1 - verif.easiness(edges, real)

        if True:
            df = pd.DataFrame({'edges': edges, 'real': real})
            df['hardness'] = hardness

            pred = verif.predict(edges)
            df['pred'] = pred.values

            df.sort_values('hardness', ascending=False)
            infr.print('hardness analysis')
            infr.print(str(df))

            infr.print('infr status: ' + ut.repr4(infr.status()))

        # Don't re-review anything that was confidently reviewed
        # CONFIDENCE = const.CONFIDENCE
        # CODE_TO_INT = CONFIDENCE.CODE_TO_INT.copy()
        # CODE_TO_INT[CONFIDENCE.CODE.UNKNOWN] = 0
        # conf = ut.take(CODE_TO_INT, infr.gen_edge_values(
        #     'confidence', edges, on_missing='default',
        #     default=CONFIDENCE.CODE.UNKNOWN))

        # This should only be run with certain params
        assert not infr.params['autoreview.enabled']
        assert not infr.params['redun.enabled']
        assert not infr.params['ranking.enabled']
        assert infr.params['inference.enabled']
        # const.CONFIDENCE.CODE.PRETTY_SURE
        if infr.params['queue.conf.thresh'] is None:
            # != 'pretty_sure':
            infr.print('WARNING: should queue.conf.thresh = "pretty_sure"?')

        # work around add_candidate_edges
        infr.prioritize(metric='hardness', edges=edges, scores=hardness)
        infr.set_edge_attrs('hardness', ut.dzip(edges, hardness))
        for _ in infr._inner_priority_gen(use_refresh=False):
            yield _

    def ranked_list_gen(infr, use_refresh=True):
        """
        Subiterator for phase1 of the main algorithm

        Calls the underlying ranking algorithm and prioritizes the results
        """
        infr.print('============================', color='white')
        infr.print('--- RANKED LIST LOOP ---', color='white')
        n_prioritized = infr.refresh_candidate_edges()
        if n_prioritized == 0:
            infr.print('RANKING ALGO FOUND NO NEW EDGES')
            return
        if use_refresh:
            infr.refresh.clear()
        for _ in infr._inner_priority_gen(use_refresh):
            yield _

    def incon_recovery_gen(infr):
        """
        Subiterator for recovery mode of the mainm algorithm

        Iterates until the graph is consistent

        Note:
            inconsistency recovery is implicitly handled by the main algorithm,
            so other phases do not need to call this explicitly. This exists
            for the case where the only mode we wish to run is inconsistency
            recovery.
        """
        maybe_error_edges = list(infr.maybe_error_edges())
        if len(maybe_error_edges) == 0:
            raise StopIteration()
        infr.print('============================', color='white')
        infr.print('--- INCON RECOVER LOOP ---', color='white')
        infr.queue.clear()
        infr.add_candidate_edges(maybe_error_edges)
        for _ in infr._inner_priority_gen(use_refresh=False):
            yield _

    def pos_redun_gen(infr):
        """
        Subiterator for phase2 of the main algorithm.

        Searches for decisions that would commplete positive redundancy

        Doctest:
            >>> from wbia.algo.graph.mixin_loops import *
            >>> import wbia
            >>> infr = wbia.AnnotInference('PZ_MTEST', aids='all',
            >>>                             autoinit='staging', verbose=4)
            >>> #infr.load_published()
            >>> gen = infr.pos_redun_gen()
            >>> feedback = next(gen)
        """
        infr.print('===========================', color='white')
        infr.print('--- POSITIVE REDUN LOOP ---', color='white')
        # FIXME: should prioritize inconsistentices first
        count = -1

        def thread_gen():
            # This is probably not safe
            new_edges = infr.find_pos_redun_candidate_edges()
            for new_edges in buffered_add_candidate_edges(infr, 50, new_edges):
                yield new_edges

        def serial_gen():
            # use this if threading does bad things
            if True:
                new_edges = list(infr.find_pos_redun_candidate_edges())
                if len(new_edges) > 0:
                    infr.add_candidate_edges(new_edges)
                    yield new_edges
            else:
                for new_edges in ub.chunks(infr.find_pos_redun_candidate_edges(), 100):
                    if len(new_edges) > 0:
                        infr.add_candidate_edges(new_edges)
                        yield new_edges

        def filtered_gen():
            # Buffer one-vs-one scores in the background and present an edge to
            # the user ASAP.
            # if infr.test_mode:
            candgen = serial_gen()
            # else:
            #     candgen = thread_gen()

            include_filter_set = None
            if PRINCETON_KAIA_EDGE_LIST is not None:
                # logger.info('[mixin_loops] FILTERING EDGES FOR KAIA')
                # Sanity check, make sure that one of the edges is in the tier 1 dataset
                include_filter_set = set(PRINCETON_KAIA_EDGE_LIST)

            for new_edges in candgen:
                if infr.ibs is not None:
                    ibs = infr.ibs
                    qual_edges = ibs.unflat_map(ibs.get_annot_quality_int, new_edges)
                    valid_edges = []
                    for (u, v), (q1, q2) in zip(new_edges, qual_edges):
                        # Skip edges involving qualities less than ok
                        if q1 is not None and q1 < ibs.const.QUAL.OK:
                            continue
                        if q2 is not None and q2 < ibs.const.QUAL.OK:
                            continue
                        if include_filter_set is not None:
                            if (
                                u not in include_filter_set
                                and v not in include_filter_set
                            ):
                                continue
                        valid_edges.append((u, v))
                    if len(valid_edges) > 0:
                        yield valid_edges
                else:
                    yield new_edges

        for count in it.count(0):
            infr.print('check pos-redun iter {}'.format(count))
            infr.queue.clear()

            found_any = False

            for new_edges in filtered_gen():
                found_any = True
                gen = infr._inner_priority_gen(use_refresh=False)
                for value in gen:
                    yield value

            # logger.info('found_any = {!r}'.format(found_any))
            if not found_any:
                break

            infr.print('not pos-reduntant yet.', color='white')
        infr.print('pos-redundancy achieved in {} iterations'.format(count + 1))

    def neg_redun_gen(infr):
        """
        Subiterator for phase3 of the main algorithm.

        Searches for decisions that would commplete negative redundancy
        """
        infr.print('===========================', color='white')
        infr.print('--- NEGATIVE REDUN LOOP ---', color='white')

        infr.queue.clear()

        only_auto = infr.params['redun.neg.only_auto']

        # TODO: outer loop that re-iterates until negative redundancy is
        # accomplished.
        needs_neg_redun = infr.find_neg_redun_candidate_edges()
        chunksize = 500
        for new_edges in ub.chunks(needs_neg_redun, chunksize):
            infr.print('another neg redun chunk')
            # Add chunks in a little at a time for faster response time
            infr.add_candidate_edges(new_edges)
            gen = infr._inner_priority_gen(use_refresh=False, only_auto=only_auto)
            for value in gen:
                yield value

    def _inner_priority_gen(infr, use_refresh=False, only_auto=False):
        """
        Helper function that implements the general inner priority loop.

        Executes reviews until the queue is empty or needs refresh

        Args:
            user_refresh (bool): if True enables the refresh criteria.
                (set to True in Phase 1)

            only_auto (bool) if True, then the user wont be prompted with
                reviews unless the graph is inconsistent.
                (set to True in Phase 3)

        Notes:
            The caller is responsible for populating the priority queue.  This
            will iterate until the queue is empty or the refresh critieron is
            triggered.
        """
        if infr.refresh:
            infr.refresh.enabled = use_refresh
        infr.print('Start inner loop with {} items in the queue'.format(len(infr.queue)))
        for count in it.count(0):
            if infr.is_recovering():
                infr.print(
                    'Still recovering after %d iterations' % (count,),
                    3,
                    color='brightcyan',
                )
            else:
                # Do not check for refresh if we are recovering
                if use_refresh and infr.refresh.check():
                    infr.print(
                        'Triggered refresh criteria after %d iterations' % (count,),
                        1,
                        color='yellow',
                    )
                    break

            # If the queue is empty break
            if len(infr.queue) == 0:
                infr.print(
                    'No more edges after %d iterations, need refresh' % (count,),
                    1,
                    color='yellow',
                )
                break

            # Try to automatically do the next review.
            edge, priority = infr.peek()
            infr.print('next_review. edge={}'.format(edge), 100)

            inconsistent = infr.is_recovering(edge)

            feedback = None
            if infr.params['autoreview.enabled'] and not inconsistent:
                # Try to autoreview if we aren't in an inconsistent state
                feedback = infr.try_auto_review(edge)

            if feedback is not None:
                # Add feedback from the automated method
                infr.add_feedback(edge, priority=priority, **feedback)
            else:
                # We can't automatically review, ask for help
                if only_auto and not inconsistent:
                    # We are in auto only mode, skip manual review
                    # unless there is an inconsistency
                    infr.skip(edge)
                else:
                    if infr.simulation_mode:
                        # Use oracle feedback
                        feedback = infr.request_oracle_review(edge)
                        infr.add_feedback(edge, priority=priority, **feedback)
                    else:
                        # Yield to the user if we need to pause
                        user_request = infr.emit_manual_review(edge, priority)
                        yield user_request

        if infr.metrics_list:
            infr._print_previous_loop_statistics(count)

    def init_refresh(infr):
        refresh_params = infr.subparams('refresh')
        infr.refresh = RefreshCriteria(**refresh_params)

    def start_id_review(infr, max_loops=None, use_refresh=None):
        assert infr._gen is None, 'algo already running'
        # Just exhaust the main generator
        infr._gen = infr.main_gen(max_loops=max_loops, use_refresh=use_refresh)
        # return infr._gen

    def main_loop(infr, max_loops=None, use_refresh=True):
        """DEPRICATED

        use list(infr.main_gen) instead
        or assert not any(infr.main_gen())
        maybe this is fine.
        """
        infr.start_id_review(max_loops=max_loops, use_refresh=use_refresh)
        # To automatically run through the loop just exhaust the generator
        try:
            result = next(infr._gen)
            assert result is None, 'need user interaction. cannot auto loop'
        except StopIteration:
            pass
        infr._gen = None


class InfrReviewers(object):
    @profile
    def try_auto_review(infr, edge):
        review = {
            'user_id': 'algo:auto_clf',
            'confidence': const.CONFIDENCE.CODE.PRETTY_SURE,
            'evidence_decision': None,
            'meta_decision': NULL,
            'timestamp_s1': None,
            'timestamp_c1': None,
            'timestamp_c2': None,
            'tags': [],
        }
        if infr.is_recovering():
            # Do not autoreview if we are in an inconsistent state
            infr.print('Must manually review inconsistent edge', 3)
            return None
        # Determine if anything passes the match threshold
        primary_task = 'match_state'

        try:
            decision_probs = infr.task_probs[primary_task][edge]
        except KeyError:
            if infr.verifiers is None:
                return None
            if infr.verifiers.get(primary_task, None) is None:
                return None
            # Compute probs if they haven't been done yet
            infr.ensure_priority_scores([edge])
            try:
                decision_probs = infr.task_probs[primary_task][edge]
            except KeyError:
                return None

        primary_thresh = infr.task_thresh[primary_task]
        decision_flags = {
            k: decision_probs[k] > thresh for k, thresh in primary_thresh.items()
        }
        hasone = sum(decision_flags.values()) == 1
        auto_flag = False
        if hasone:
            try:
                # Check to see if it might be confounded by a photobomb
                pb_probs = infr.task_probs['photobomb_state'][edge]
                # pb_probs = infr.task_probs['photobomb_state'].loc[edge]
                # pb_probs = data['task_probs']['photobomb_state']
                pb_thresh = infr.task_thresh['photobomb_state']['pb']
                confounded = pb_probs['pb'] > pb_thresh
            except KeyError:
                logger.info('Warning: confounding task probs not set (i.e. photobombs)')
                confounded = False
            if not confounded:
                # decision = decision_flags.argmax()
                evidence_decision = ut.argmax(decision_probs)
                review['evidence_decision'] = evidence_decision
                truth = infr.match_state_gt(edge)
                if review['evidence_decision'] != truth:
                    infr.print(
                        'AUTOMATIC ERROR edge={}, truth={}, decision={}, probs={}'.format(
                            edge, truth, review['evidence_decision'], decision_probs
                        ),
                        2,
                        color='red',
                    )
                auto_flag = True
        if auto_flag and infr.verbose > 1:
            infr.print('Automatic review success')

        if auto_flag:
            return review
        else:
            return None

    def request_oracle_review(infr, edge, **kw):
        truth = infr.match_state_gt(edge)
        feedback = infr.oracle.review(edge, truth, infr, **kw)
        return feedback

    def _make_review_tuple(infr, edge, priority=None):
        """ Makes tuple to be sent back to the user """
        edge_data = infr.get_nonvisual_edge_data(edge, on_missing='default')
        # Extra information
        edge_data['nid_edge'] = infr.pos_graph.node_labels(*edge)
        if infr.queue is None:
            edge_data['queue_len'] = 0
        else:
            edge_data['queue_len'] = len(infr.queue)
        edge_data['n_ccs'] = (
            len(infr.pos_graph.connected_to(edge[0])),
            len(infr.pos_graph.connected_to(edge[1])),
        )
        return (edge, priority, edge_data)

    def emit_manual_review(infr, edge, priority=None):
        """
        Emits a signal containing edges that need review. The callback should
        present them to a user, get feedback, and then call on_accpet.
        """
        infr.print('emit_manual_review', 100)
        # Emit a list of reviews that can be considered.
        # The first is the most important
        user_request = []
        user_request += [infr._make_review_tuple(edge, priority)]
        try:
            for edge_, priority in infr.peek_many(infr.params['manual.n_peek']):
                if edge == edge_:
                    continue
                user_request += [infr._make_review_tuple(edge_, priority)]
        except TypeError:
            pass

        # If registered, send the request via a callback.
        request_review = infr.callbacks.get('request_review', None)
        if request_review is not None:
            # Send these reviews to a user
            request_review(user_request)
        # Otherwise the current process must handle the request by return value
        return user_request

    def skip(infr, edge):
        infr.print('skipping edge={}'.format(edge), 100)
        try:
            del infr.queue[edge]
        except Exception:
            pass

    def accept(infr, feedback):
        """
        Called when user has completed feedback from qt or web
        """
        annot1_state = feedback.pop('annot1_state', None)
        annot2_state = feedback.pop('annot2_state', None)
        if annot1_state:
            infr.add_node_feedback(**annot1_state)
        if annot2_state:
            infr.add_node_feedback(**annot2_state)
        infr.add_feedback(**feedback)

        if infr.params['manual.autosave']:
            infr.write_wbia_staging_feedback()

    def continue_review(infr):
        infr.print('continue_review', 10)
        if infr._gen is None:
            return None
        try:
            user_request = next(infr._gen)
        except StopIteration:
            review_finished = infr.callbacks.get('review_finished', None)
            if review_finished is not None:
                review_finished()
            infr._gen = None
            user_request = None
        return user_request

    def qt_edge_reviewer(infr, edge=None):
        import wbia.guitool as gt

        gt.ensure_qapp()
        from wbia.viz import viz_graph2

        infr.manual_wgt = viz_graph2.AnnotPairDialog(
            edge=edge, infr=infr, standalone=False, cfgdict=infr.verifier_params
        )
        if edge is not None:
            # infr.emit_manual_review(edge, priority=None)
            infr.manual_wgt.seek(0)
            # infr.manual_wgt.show()
        return infr.manual_wgt

    def qt_review_loop(infr):
        r"""
        TODO: The loop parts should be a non-mixin class

        Qt review loop entry point

        CommandLine:
            python -m wbia.algo.graph.mixin_loops qt_review_loop --show

        Example:
            >>> # SCRIPT
            >>> import utool as ut
            >>> import wbia
            >>> ibs = wbia.opendb('PZ_MTEST')
            >>> infr = wbia.AnnotInference(ibs, 'all', autoinit=True)
            >>> infr.ensure_mst()
            >>> # Add dummy priorities to each edge
            >>> infr.set_edge_attrs('prob_match', ut.dzip(infr.edges(), [1]))
            >>> infr.prioritize('prob_match', infr.edges(), reset=True)
            >>> infr.params['redun.enabled'] = False
            >>> win = infr.qt_review_loop()
            >>> import wbia.guitool as gt
            >>> gt.qtapp_loop(qwin=win, freq=10)
        """
        infr.qt_edge_reviewer()
        # infr.continue_review()
        return infr.manual_wgt


if False:
    # Testing generating using threads
    from threading import Thread

    _sentinel = object()

    class _background_consumer(Thread):
        """
        Will fill the queue with content of the source in a separate thread.

        Ignore:
            >>> from wbia.algo.graph.mixin_loops import *
            >>> import wbia
            >>> infr = wbia.AnnotInference('PZ_MTEST', aids='all',
            >>>                             autoinit='staging', verbose=4)
            >>> infr.load_published()
            >>> gen = infr.find_pos_redun_candidate_edges()
            >>> parbuf = buffered_add_candidate_edges(infr, 3, gen)
            >>> next(parbuf)

        """

        def __init__(self, infr, queue, source):
            Thread.__init__(self)

            self.infr = infr

            self._queue = queue
            self._source = source

        def run(self):
            # for edges in ub.chunks(self._source, 5):
            #     logger.info('edges = {!r}'.format(edges))
            #     # logger.info('put item = {!r}'.format(item))
            #     # probably not thread safe
            #     infr = self.infr
            #     infr.add_candidate_edges(edges)
            #     for item in edges:
            #         self._queue.put(item)
            for _, item in enumerate(self._source):
                # import threading
                # import multiprocessing
                # logger.info('multiproc = ' + str(multiprocessing.current_process()))
                # logger.info('thread = ' + str(threading.current_thread()))
                # logger.info('_ = {!r}'.format(_))
                # logger.info('item = {!r}'.format(item))
                # logger.info('put item = {!r}'.format(item))
                # probably not thread safe
                infr = self.infr
                infr.add_candidate_edges([item])
                self._queue.put(item, block=True)

            # Signal the consumer we are done.
            self._queue.put(_sentinel)

    class buffered_add_candidate_edges(object):
        """
        Buffers content of an iterator polling the contents of the given
        iterator in a separate thread.
        When the consumer is faster than many producers, this kind of
        concurrency and buffering makes sense.

        The size parameter is the number of elements to buffer.

        The source must be threadsafe.
        """

        def __init__(self, infr, size, source):
            if six.PY2:
                from Queue import Queue
            else:
                from queue import Queue
            self._queue = Queue(size)

            self._poller = _background_consumer(infr, self._queue, source)
            self._poller.daemon = True
            self._poller.start()

        def __iter__(self):
            return self

        def __next__(self):
            item = self._queue.get(True)
            if item is _sentinel:
                raise StopIteration()
            return item

        next = __next__



# -*- coding: utf-8 -*-
"""
Matplotlib interface for name interactions. Allows for relatively fine grained
control of splitting and merging.

DEPRICATE

CommandLine:
    python -m wbia.viz.interact.interact_name --test-ishow_name --show
    python -m wbia.viz.interact.interact_name --test-testsdata_match_verification --show --db PZ_MTEST --aid1 1 --aid2 30
    python -m wbia.viz.interact.interact_name --test-testsdata_match_verification --show --db PZ_MTEST --aid1 30 --aid2 32


"""
import logging
import numpy as np
import utool as ut
from six.moves import zip
from wbia.plottool import interact_helpers as ih
import functools
import wbia.plottool as pt
from wbia import viz
from wbia import constants as const
from wbia.viz import viz_helpers as vh
from wbia.other import ibsfuncs
from wbia.viz import viz_chip
from wbia.plottool.abstract_interaction import AbstractInteraction

(print, rrr, profile) = ut.inject2(__name__, '[interact_name]', DEBUG=False)
logger = logging.getLogger('wbia')


# ==========================
# Name Interaction
# ==========================

MAX_COLS = 3


def build_name_context_options(ibs, nids):
    logger.info('build_name_context_options nids = %r' % (nids,))
    callback_list = []
    from wbia.viz import viz_graph2

    callback_list.extend(
        [
            (
                'New Split Interact (Name)',
                functools.partial(viz_graph2.make_qt_graph_interface, ibs, nids=nids),
            ),
        ]
    )
    return callback_list


def ishow_name(
    ibs, nid, sel_aids=[], select_aid_callback=None, fnum=5, dodraw=True, **kwargs
):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        nid (?):
        sel_aids (list):
        select_aid_callback (None):
        fnum (int):  figure number

    CommandLine:
        python -m wbia.viz.interact.interact_name --test-ishow_name --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.viz.interact.interact_name import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> nid = ut.get_argval('--nid', int, default=1)
        >>> sel_aids = []
        >>> select_aid_callback = None
        >>> fnum = 5
        >>> dodraw = ut.show_was_requested()
        >>> # execute function
        >>> result = ishow_name(ibs, nid, sel_aids, select_aid_callback, fnum, dodraw)
        >>> # verify results
        >>> pt.show_if_requested()
        >>> print(result)
    """
    if fnum is None:
        fnum = pt.next_fnum()
    fig = ih.begin_interaction('name', fnum)

    def _on_name_click(event):
        ax = event.inaxes
        if ih.clicked_inside_axis(event):
            viztype = vh.get_ibsdat(ax, 'viztype')
            if viztype == 'chip':
                aid = vh.get_ibsdat(ax, 'aid')
                logger.info('... aid=%r' % aid)
                if event.button == 3:  # right-click
                    from wbia import guitool
                    from wbia.viz.interact import interact_chip

                    height = fig.canvas.geometry().height()
                    qpoint = guitool.newQPoint(event.x, height - event.y)
                    refresh_func = functools.partial(
                        viz.show_name, ibs, nid, fnum=fnum, sel_aids=sel_aids
                    )
                    interact_chip.show_annot_context_menu(
                        ibs,
                        aid,
                        fig.canvas,
                        qpoint,
                        refresh_func=refresh_func,
                        with_interact_name=False,
                    )
                else:
                    viz.show_name(ibs, nid, fnum=fnum, sel_aids=[aid], in_image=True)
                    if select_aid_callback is not None:
                        select_aid_callback(aid)
        viz.draw()

    viz.show_name(ibs, nid, fnum=fnum, sel_aids=sel_aids, in_image=True)
    if dodraw:
        viz.draw()
    ih.connect_callback(fig, 'button_press_event', _on_name_click)
    pass


def testsdata_match_verification(defaultdb='testdb1', aid1=1, aid2=2):
    r"""
    CommandLine:
        main.py --imgsetid 2
        main.py --imgsetid 13 --db PZ_MUGU_19

    CommandLine:
        python -m wbia.viz.interact.interact_name --test-testsdata_match_verification --show
        python -m wbia.viz.interact.interact_name --test-testsdata_match_verification --aid1 2 --aid2 3 --show

        # Merge case
        python -m wbia.viz.interact.interact_name --test-testsdata_match_verification --show --db PZ_MTEST --aid1 1 --aid2 30

        # Split case
        python -m wbia.viz.interact.interact_name --test-testsdata_match_verification --show --db PZ_MTEST --aid1 30 --aid2 32

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.viz.interact.interact_name import *  # NOQA
        >>> self = testsdata_match_verification()
        >>> # verify results
        >>> ut.quit_if_noshow()
        >>> self.show_page()
        >>> ut.show_if_requested()
    """
    # from wbia.viz.interact.interact_name import *  # NOQA
    import wbia

    # ibs = wbia.opendb(defaultdb='PZ_Master0')
    ibs = wbia.opendb(defaultdb=defaultdb)
    # aid1 = ut.get_argval('--aid1', int, 14)
    # aid2 = ut.get_argval('--aid2', int, 5545)
    aid1 = ut.get_argval('--aid1', int, aid1)
    aid2 = ut.get_argval('--aid2', int, aid2)
    self = MatchVerificationInteraction(ibs, aid1, aid2, dodraw=False)
    return self


class MatchVerificationInteraction(AbstractInteraction):
    def __init__(
        self,
        ibs,
        aid1,
        aid2,
        update_callback=None,
        backend_callback=None,
        dodraw=True,
        max_cols=MAX_COLS,
        **kwargs,
    ):
        if ut.VERBOSE:
            logger.info('[matchver] __init__')
        if ut.VERBOSE or ut.is_developer():
            logger.info('[matchver] __init__ aid1=%r, aid2=%r ' % (aid1, aid2))
        super(MatchVerificationInteraction, self).__init__(**kwargs)
        self.ibs = ibs
        self.max_cols = max_cols
        self.aid1 = aid1
        self.aid2 = aid2
        self.col_offset_list = [0, 0]
        # ibsfuncs.assert_valid_aids(ibs, [aid1, aid2])

        def _nonefn():
            return None

        if update_callback is None:
            update_callback = _nonefn
        if backend_callback is None:
            backend_callback = _nonefn
        self.update_callback = (
            update_callback  # if something like qt needs a manual refresh on change
        )
        self.backend_callback = backend_callback
        self.qres_callback = kwargs.get('qres_callback', None)
        self.cm = kwargs.get('cm', None)
        self.qreq_ = kwargs.get('qreq_', None)
        if self.cm is not None:
            from wbia.algo.hots import chip_match

            assert isinstance(self.cm, chip_match.ChipMatch)
            assert self.qreq_ is not None
        self.infer_data()
        if dodraw:
            self.show_page(bring_to_front=True)

    def infer_data(self):
        """Initialize data related to the input aids"""
        ibs = self.ibs
        # The two matching aids
        self.aid_pair = (self.aid1, self.aid2)
        (aid1, aid2) = self.aid_pair
        self.match_text = ibs.get_match_text(self.aid1, self.aid2)
        # The names of the matching annotations
        self.nid1, self.nid2 = ibs.get_annot_name_rowids((aid1, aid2))
        self.name1, self.name2 = ibs.get_annot_names((aid1, aid2))
        self.other_valid_nids = []
        # The other annotations that belong to these two names
        self.gts_list = ibs.get_annot_groundtruth((aid1, aid2))
        self.gt1, self.gt2 = self.gts_list
        # A flat list of all the aids we are looking at
        self.is_split_case = self.nid1 == self.nid2
        self.all_aid_list = ut.unique_ordered([aid1, aid2] + self.gt1 + self.gt2)
        self.all_nid_list_orig = ibs.get_annot_name_rowids(self.all_aid_list)
        self.other_aids = list(set(self.all_aid_list) - set([self.aid1, self.aid2]))

        if self.is_split_case:
            # Split case
            self.nCols = max(2, len(self.other_aids))
            self.nRows = 2 if len(self.other_aids) > 0 else 1
        else:
            # Merge/New Match case
            self.nCols = max(len(self.gt1) + 1, len(self.gt2) + 1)
            self.nRows = 2
        self.nCols = min(self.max_cols, self.nCols)

        # Grab not just the exemplars

        if ut.VERBOSE or ut.is_developer():
            logger.info('[matchver] __init__ nid1=%r, nid2=%r ' % (self.nid1, self.nid2))
            logger.info('[matchver] __init__ self.gts_list=%r ' % (self.gts_list))

        if ut.VERBOSE or ut.is_developer():
            logger.info('[matchver] __init__ nid1=%r, nid2=%r ' % (self.nid1, self.nid2))
            logger.info('[matchver] __init__ self.gts_list=%r ' % (self.gts_list))

    def get_other_nids(self):
        ibs = self.ibs
        all_nid_list = ibs.get_annot_name_rowids(self.all_aid_list)
        unique_nid_list = ut.unique_ordered(all_nid_list)
        is_unknown = ibs.is_nid_unknown(unique_nid_list)
        is_name1 = [nid == self.nid1 for nid in unique_nid_list]
        is_name2 = [nid == self.nid2 for nid in unique_nid_list]
        is_other = ut.and_lists(
            *tuple(map(ut.not_list, (is_name1, is_name2, is_unknown)))
        )
        other_nid_list = ut.compress(unique_nid_list, is_other)
        return other_nid_list

    def get_rotating_columns(self, rowx):
        if self.is_split_case:
            if rowx == 0:
                return []
            else:
                return self.other_aids
        else:
            if rowx == 0:
                return self.gt1
            else:
                return self.gt2

    def get_non_rotating_columns(self, rowx):
        if self.is_split_case:
            if rowx == 0:
                return [self.aid1, self.aid2]
            else:
                return []
        else:
            if rowx == 0:
                return [self.aid1]
            else:
                return [self.aid2]

    def get_row_aids_list(self):
        r"""
        Args:

        Returns:
            list: row_aids_list

        CommandLine:
            python -m wbia.viz.interact.interact_name --test-get_row_aids_list

        CommandLine:
            python -m wbia.viz.interact.interact_name --test-get_row_aids_list
            python -m wbia.viz.interact.interact_name --test-get_row_aids_list --aid1 2 --aid2 3
            # Merge case
            python -m wbia.viz.interact.interact_name --test-get_row_aids_list --db PZ_MTEST --aid1 1 --aid2 30
            # Split case
            python -m wbia.viz.interact.interact_name --test-get_row_aids_list --db PZ_MTEST --aid1 30 --aid2 32

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.viz.interact.interact_name import *  # NOQA
            >>> # build test data
            >>> self = testsdata_match_verification('PZ_MTEST', 30, 32)
            >>> # execute function
            >>> row_aids_list = self.get_row_aids_list()
            >>> # verify results
            >>> result = str(row_aids_list)
            >>> print(result)
            >>> ut.quit_if_noshow()
            >>> self.show_page()
            >>> ut.show_if_requested()
        """

        def get_row(rowx):
            row_offset = self.col_offset_list[rowx]
            row_nonrotate_part = self.get_non_rotating_columns(rowx)
            row_rotate_part_ = self.get_rotating_columns(rowx)
            row_rotate_part = ut.list_roll(row_rotate_part_, -row_offset)
            row = row_nonrotate_part + row_rotate_part
            return row

        row_aids_list_ = [get_row(rowx) for rowx in range(self.nRows)]
        row_aids_list = list(filter(lambda x: len(x) > 0, row_aids_list_))
        return row_aids_list

    def rotate_row(self, event=None, rowx=None):
        """
        shows the next few annotations in this row
        (implicitly rotates the row's columns the rows columns)
        """
        modbase = len(self.get_rotating_columns(rowx))
        self.col_offset_list[rowx] = (1 + self.col_offset_list[rowx]) % modbase
        # self.gts_list[rowx] = list_roll(self.gts_list[rowx], -(self.nCols - 1))
        self.show_page(onlyrows=[rowx], fulldraw=False)

    def prepare_page(self, fulldraw=True):
        figkw = {
            'fnum': self.fnum,
            'doclf': fulldraw,
            'docla': fulldraw,
        }
        if fulldraw:
            self.fig = pt.figure(**figkw)
        ih.disconnect_callback(self.fig, 'button_press_event')
        ih.disconnect_callback(self.fig, 'key_press_event')
        ih.connect_callback(self.fig, 'button_press_event', self.figure_clicked)
        ih.connect_callback(self.fig, 'key_press_event', self.on_key_press)

    def show_page(self, bring_to_front=False, onlyrows=None, fulldraw=True):
        """Plots all subaxes on a page

        onlyrows is a hack to only draw a subset of the data again
        """
        if ut.VERBOSE:
            if not fulldraw:
                logger.info(
                    '[matchver] show_page(fulldraw=%r, onlyrows=%r)'
                    % (fulldraw, onlyrows)
                )
            else:
                logger.info('[matchver] show_page(fulldraw=%r)' % (fulldraw))
        self.prepare_page(fulldraw=fulldraw)
        # Variables we will work with to paint a pretty picture
        ibs = self.ibs
        nRows = self.nRows
        colpad = 1 if self.cm is not None else 0
        nCols = self.nCols + colpad

        # Distinct color for every unique name
        unique_nids = ut.unique_ordered(
            ibs.get_annot_name_rowids(self.all_aid_list, distinguish_unknowns=False)
        )
        unique_colors = pt.distinct_colors(
            len(unique_nids), brightness=0.7, hue_range=(0.05, 0.95)
        )
        self.nid2_color = dict(zip(unique_nids, unique_colors))

        row_aids_list = self.get_row_aids_list()

        if self.cm is not None:
            logger.info('DRAWING QRES')
            pnum = (1, nCols, 1)
            if not fulldraw:
                # not doing full draw so we have to clear any axes
                # that are here already manually
                ax = self.fig.add_subplot(*pnum)
                self.clear_parent_axes(ax)
            self.cm.show_single_annotmatch(
                self.qreq_,
                self.aid2,
                fnum=self.fnum,
                pnum=pnum,
                draw_fmatch=True,
                colorbar_=False,
            )

        # For each row
        for rowx, aid_list in enumerate(row_aids_list):
            offset = rowx * nCols + 1
            if onlyrows is not None and rowx not in onlyrows:
                continue
            # ibsfuncs.assert_valid_aids(ibs, groundtruth)
            # For each column
            for colx, aid in enumerate(aid_list, start=colpad):
                if colx >= nCols:
                    break
                try:
                    nid = ibs.get_annot_name_rowids(aid)
                    if ibsfuncs.is_nid_unknown(ibs, [nid])[0]:
                        color = const.UNKNOWN_PURPLE_RGBA01
                    else:
                        color = self.nid2_color[nid]
                except Exception as ex:
                    ut.printex(ex)
                    logger.info('nid = %r' % (nid,))
                    logger.info('self.nid2_color = %s' % (ut.repr2(self.nid2_color),))
                    raise
                px = colx + offset
                ax = self.plot_chip(
                    int(aid), nRows, nCols, px, color=color, fulldraw=fulldraw
                )
                # If there are still more in this row to display
                if colx + 1 < len(aid_list) and colx + 1 >= nCols:
                    total_indices = len(aid_list)
                    current_index = self.col_offset_list[rowx] + 1
                    next_text = 'next\n%d/%d' % (current_index, total_indices)
                    next_func = functools.partial(self.rotate_row, rowx=rowx)
                    self.append_button(
                        next_text,
                        callback=next_func,
                        location='right',
                        size='33%',
                        ax=ax,
                    )

        if fulldraw:
            self.show_hud()
            hspace = 0.05 if (self.nCols) > 1 else 0.1
            subplotspar = {
                'left': 0.1,
                'right': 0.9,
                'top': 0.85,
                'bottom': 0.1,
                'wspace': 0.3,
                'hspace': hspace,
            }
            pt.adjust_subplots(**subplotspar)
        self.draw()
        self.show()
        if bring_to_front:
            self.bring_to_front()
        # self.update()

    def plot_chip(self, aid, nRows, nCols, px, fulldraw=True, **kwargs):
        """ Plots an individual chip in a subaxis """
        ibs = self.ibs
        if aid in [self.aid1, self.aid2]:
            # Bold color for the matching chips
            lw = 5
            text_color = np.array((135, 206, 235, 255)) / 255.0
        else:
            lw = 2
            text_color = None

        pnum = (nRows, nCols, px)
        if not fulldraw:
            # not doing full draw so we have to clear any axes
            # that are here already manually
            ax = self.fig.add_subplot(*pnum)
            self.clear_parent_axes(ax)
            # ut.embed()
            # logger.info(subax)

        viz_chip_kw = {
            'fnum': self.fnum,
            'pnum': pnum,
            'nokpts': True,
            'show_name': True,
            'show_gname': False,
            'show_aidstr': True,
            'notitle': True,
            'show_num_gt': False,
            'text_color': text_color,
        }
        if False and ut.is_developer():
            enable_chip_title_prefix = True
            viz_chip_kw.update(
                {
                    'enable_chip_title_prefix': enable_chip_title_prefix,
                    'show_name': True,
                    'show_aidstr': True,
                    'show_viewcode': True,
                    'show_num_gt': True,
                    'show_quality_text': True,
                }
            )

        viz_chip.show_chip(ibs, aid, **viz_chip_kw)
        ax = pt.gca()
        pt.draw_border(ax, color=kwargs.get('color'), lw=lw)
        if kwargs.get('make_buttons', True):
            # divider = pt.ensure_divider(ax)
            butkw = {
                # 'divider': divider,
                'ax': ax,
                'size': '13%'
                # 'size': '15%'
            }
        # Chip matching/naming options
        nid = ibs.get_annot_name_rowids(aid)
        annotation_unknown = ibs.is_nid_unknown([nid])[0]
        if not annotation_unknown:
            # remove name
            callback = functools.partial(self.unname_annotation, aid)
            self.append_button(
                'remove name (' + ibs.get_name_texts(nid) + ')',
                callback=callback,
                **butkw,
            )
        else:
            # new name
            callback = functools.partial(self.mark_annotation_as_new_name, aid)
            self.append_button('mark as new name', callback=callback, **butkw)
        if (
            nid != self.nid2
            and not ibs.is_nid_unknown([self.nid2])[0]
            and not self.is_split_case
        ):
            # match to nid2
            callback = functools.partial(self.rename_annotation, aid, self.nid2)
            text = 'match to name2: ' + ibs.get_name_texts(self.nid2)
            self.append_button(text, callback=callback, **butkw)
        if nid != self.nid1 and not ibs.is_nid_unknown([self.nid1])[0]:
            # match to nid1
            callback = functools.partial(self.rename_annotation, aid, self.nid1)
            text = 'match to name1: ' + ibs.get_name_texts(self.nid1)
            self.append_button(text, callback=callback, **butkw)

        other_nid_list = self.get_other_nids()
        for other_nid in other_nid_list:
            if other_nid == nid:
                continue
            # rename nid2
            callback = functools.partial(self.rename_annotation, aid, other_nid)
            text = 'match to: ' + ibs.get_name_texts(other_nid)
            self.append_button(text, callback=callback, **butkw)
        return ax

    def show_hud(self):
        """Creates heads up display

        button bar on bottom and title string

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.viz.interact.interact_name import *  # NOQA
            >>> # build test data
            >>> self = testsdata_match_verification('PZ_MTEST', 30, 32)
            >>> # execute function
            >>> result = self.show_hud()
            >>> # verify results
            >>> print(result)
            >>> ut.quit_if_noshow()
            >>> self.show_page()
            >>> pt.show_if_requested()
        """
        # Button positioners
        hl_slot, hr_slot = pt.make_bbox_positioners(
            y=0.02, w=0.15, h=0.063, xpad=0.02, startx=0, stopx=1
        )
        # hack make a second bbox positioner to get different sized buttons on #
        # the left
        hl_slot2, hr_slot2 = pt.make_bbox_positioners(
            y=0.02, w=0.08, h=0.05, xpad=0.015, startx=0, stopx=1
        )

        def next_rect(accum=[-1]):
            accum[0] += 1
            return hr_slot(accum[0])

        def next_rect2(accum=[-1]):
            accum[0] += 1
            return hl_slot2(accum[0])

        ibs = self.ibs
        name1, name2 = self.name1, self.name2
        nid1_is_known = not ibs.is_nid_unknown(self.nid1)
        nid2_is_known = not ibs.is_nid_unknown(self.nid2)
        all_nid_list = ibs.get_annot_name_rowids(self.all_aid_list)
        is_unknown = ibs.is_nid_unknown(all_nid_list)
        is_name1 = [nid == self.nid1 for nid in all_nid_list]
        is_name2 = [nid == self.nid2 for nid in all_nid_list]

        # option to remove all names only if at least one name exists
        if not all(is_unknown):
            unname_all_text = 'remove all names'
            self.append_button(
                unname_all_text, callback=self.unname_all, rect=next_rect()
            )
        # option to merge all into a new name if all are unknown
        if all(is_unknown) and not nid1_is_known and not nid2_is_known:
            joinnew_text = 'match all (nonjunk)\n to a new name'
            self.append_button(
                joinnew_text, callback=self.merge_nonjunk_into_new_name, rect=next_rect()
            )
        # option dismiss all and give new names to all nonjunk images
        if any(is_unknown):
            self.append_button(
                'mark all unknowns\nas not matching',
                callback=self.dismiss_all,
                rect=next_rect(),
            )
        # merges all into the first name
        if nid1_is_known and not all(is_name1):
            join1_text = 'match all to name1:\n{name1}'.format(name1=name1)
            callback = functools.partial(self.merge_all_into_nid, self.nid1)
            self.append_button(join1_text, callback=callback, rect=next_rect())
        # merges all into the seoncd name
        if name1 != name2 and nid2_is_known and not all(is_name2):
            join2_text = 'match all to name2:\n{name2}'.format(name2=name2)
            callback = functools.partial(self.merge_all_into_nid, self.nid2)
            self.append_button(join2_text, callback=callback, rect=next_rect())
        ###
        self.append_button('close', callback=self.close_, rect=next_rect2())
        if self.qres_callback is not None:
            self.append_button('review', callback=self.review, rect=next_rect2())
        self.append_button('reset', callback=self.reset_all_names, rect=next_rect2())
        self.dbname = ibs.get_dbname()
        self.vsstr = 'qaid%d-vs-aid%d' % (self.aid1, self.aid2)
        figtitle_fmt = """
        Match Review Interface - {dbname}
        {match_text}:
        {vsstr}
        """
        figtitle = figtitle_fmt.format(**self.__dict__)  # sexy: using obj dict as fmtkw
        pt.set_figtitle(figtitle)

    def on_close(self, event=None):
        super(MatchVerificationInteraction, self).on_close(event)
        pass

    def unname_annotation(self, aid, event=None):
        if ut.VERBOSE:
            logger.info('remove name')
        self.ibs.delete_annot_nids([aid])
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def mark_annotation_as_new_name(self, aid, event=None):
        if ut.VERBOSE:
            logger.info('new name')
        self.ibs.set_annot_names_to_same_new_name([aid])
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def rename_annotation(self, aid, nid, event=None):
        if ut.VERBOSE:
            logger.info('rename nid1')
        self.ibs.set_annot_name_rowids([aid], [nid])
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def reset_all_names(self, event=None):
        self.ibs.set_annot_name_rowids(self.all_aid_list, self.all_nid_list_orig)
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def review(self, event=None):
        if ut.VERBOSE:
            logger.info('review pressed')
        if self.qres_callback is not None:
            self.qres_callback()
        else:
            logger.info('Warning: no review callback connected.')

    def close_(self, event=None):
        # closing this gui with the button means you have reviewed the annotation.
        # self.ibs.set_annot_pair_as_reviewed(self.aid1, self.aid2)
        self.close()

    def unname_all(self, event=None):
        if ut.VERBOSE:
            logger.info('unname_all')
        self.ibs.delete_annot_nids(self.all_aid_list)
        self.show_page()

    def merge_all_into_nid(self, nid, event=None):
        """ All the annotations are given nid """
        aid_list = self.all_aid_list
        self.ibs.set_annot_name_rowids(aid_list, [nid] * len(aid_list))
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def merge_nonjunk_into_new_name(self, event=None):
        """ All nonjunk annotations are given the SAME new name """
        # Delete all original names
        aid_list = self.all_aid_list
        aid_list_filtered = ut.filterfalse_items(
            aid_list, self.ibs.get_annot_isjunk(aid_list)
        )
        # Rename annotations
        self.ibs.set_annot_names_to_same_new_name(aid_list_filtered)
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def dismiss_all(self, event=None):
        """ All unknown annotations are given DIFFERENT new names """
        # Delete all original names
        ibs = self.ibs
        aid_list = self.all_aid_list
        is_unknown = ibs.is_aid_unknown(aid_list)
        aid_list_filtered = ut.compress(aid_list, is_unknown)
        # Rename annotations
        ibs.set_annot_names_to_different_new_names(aid_list_filtered)
        self.update_callback()
        self.backend_callback()
        self.show_page()

    def on_key_press(self, event=None):
        if event.key == 'escape':
            from wbia import guitool

            if guitool.are_you_sure():
                self.close()

    def figure_clicked(self, event=None):
        ax = event.inaxes
        if ih.clicked_inside_axis(event):
            viztype = vh.get_ibsdat(ax, 'viztype')
            if viztype == 'chip':
                aid = vh.get_ibsdat(ax, 'aid')
                # logger.info('... aid=%r' % aid)
                if event.button == 3:  # right-click
                    # import wbia.guitool
                    # height = self.fig.canvas.geometry().height()
                    # qpoint = guitool.newQPoint(event.x, height - event.y)
                    # ibs = self.ibs
                    # is_exemplar = ibs.get_annot_exemplar_flags(aid)
                    # def context_func():
                    #    ibs.set_annot_exemplar_flags(aid, not is_exemplar)
                    #    self.show_page()
                    # guitool.popup_menu(self.fig.canvas, pt, [
                    #    ('unset as exemplar' if is_exemplar else 'set as exemplar', context_func),
                    # ])
                    # TODO USE ABSTRACT INTERACTION
                    from wbia.viz.interact import interact_chip

                    options = interact_chip.build_annot_context_options(
                        self.ibs, aid, refresh_func=self.show_page
                    )
                    self.show_popup_menu(options, event)
                    # interact_chip.show_annot_context_menu(
                    #    self.ibs, aid, self.fig.canvas, qpoint, refresh_func=self.show_page)
                    # ibs.print_annotation_table()
                # logger.info(ut.repr2(event.__dict__))
            elif viztype == 'matches':
                self.cm.ishow_single_annotmatch(self.qreq_, self.aid2, fnum=None, mode=0)



# -*- coding: utf-8 -*-
import logging
import wbia.plottool as pt
import utool as ut
import numpy as np
from wbia.other import ibsfuncs
from wbia.viz import viz_helpers as vh
from wbia.viz import viz_chip
from wbia.viz import viz_matches  # NOQA

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


DEFAULT_NTOP = 3


def show_qres_top(ibs, cm, qreq_=None, **kwargs):
    """
    Wrapper around show_qres.
    """
    N = kwargs.get('N', DEFAULT_NTOP)
    # name_scoring = kwargs.get('name_scoring', False)
    # if isinstance(cm, chip_match.ChipMatch):
    top_aids = cm.get_top_aids(N)
    # else:
    #    top_aids = cm.get_top_aids(num=N, ibs=ibs, name_scoring=name_scoring)
    aidstr = ibsfuncs.aidstr(cm.qaid)
    figtitle = kwargs.get('figtitle', '')
    if len(figtitle) > 0:
        figtitle = ' ' + figtitle
    kwargs['figtitle'] = ('q%s -- TOP %r' % (aidstr, N)) + figtitle
    return show_qres(
        ibs,
        cm,
        top_aids=top_aids,
        qreq_=qreq_,
        # dont use these. use annot mode instead
        # draw_kpts=False,
        # draw_ell=False,
        # all_kpts=False,
        **kwargs,
    )


def show_qres_analysis(ibs, cm, qreq_=None, **kwargs):
    """
    Wrapper around show_qres.

    KWARGS:
        aid_list - show matches against aid_list (default top 3)

    Args:
        ibs (IBEISController):  wbia controller object
        cm (ChipMatch):  object of feature correspondences and scores
        qreq_ (QueryRequest):  query request object with hyper-parameters(default = None)

    Kwargs:
        N, show_gt, show_query, aid_list, figtitle, viz_name_score, viz_name_score

    CommandLine:
        python -m wbia.viz.viz_qres --exec-show_qres_analysis --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.viz.viz_qres import *  # NOQA
        >>> import wbia
        >>> cm, qreq_ = wbia.testdata_cm(
        >>>     defaultdb='PZ_MTEST', default_qaids=[1],
        >>>     default_daids=[2, 3, 4, 5, 6, 7, 8, 9])
        >>> kwargs = dict(show_query=False, viz_name_score=True,
        >>>               show_timedelta=True, N=3, show_gf=True)
        >>> ibs = qreq_.ibs
        >>> show_qres_analysis(ibs, cm, qreq_, **kwargs)
        >>> ut.show_if_requested()

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.viz.viz_qres import *  # NOQA
        >>> import wbia
        >>> cm, qreq_ = wbia.testdata_cm(
        >>>     defaultdb='PZ_MTEST', default_qaids=[1],
        >>>     default_daids=[2])
        >>> kwargs = dict(show_query=False, viz_name_score=True,
        >>>               show_timedelta=True, N=3, show_gf=True)
        >>> ibs = qreq_.ibs
        >>> show_qres_analysis(ibs, cm, qreq_, **kwargs)
        >>> ut.show_if_requested()
    """
    if ut.NOT_QUIET:
        logger.info('[show_qres] cm.show_analysis()')
    # Parse arguments
    N = kwargs.get('N', DEFAULT_NTOP)
    show_gt = kwargs.pop('show_gt', True)
    show_gf = kwargs.pop('show_gf', False)
    show_query = kwargs.pop('show_query', True)
    aid_list = kwargs.pop('aid_list', None)
    figtitle = kwargs.pop('figtitle', None)
    viz_name_score = kwargs.get('viz_name_score', True)
    failed_to_match = False

    if aid_list is None:
        # Compare to aid_list instead of using top ranks
        top_aids = cm.get_top_aids(N)
        if len(top_aids) == 0:
            failed_to_match = True
            logger.info('WARNING! No matches found for this query')
        if figtitle is None:
            if len(top_aids) == 0:
                figtitle = 'WARNING: no matches found!' + ibsfuncs.aidstr(cm.qaid)
            else:
                topscore = cm.get_annot_scores(top_aids)[0]
                figtitle = 'q%s -- topscore=%r' % (ibsfuncs.aidstr(cm.qaid), topscore)
    else:
        logger.info('[analysis] showing a given list of aids')
        top_aids = aid_list
        if figtitle is None:
            figtitle = 'comparing to ' + ibsfuncs.aidstr(top_aids) + figtitle

    # Get any groundtruth if you are showing it
    showgt_aids = []
    if show_gt:
        # Get the missed groundtruth annotations
        assert qreq_ is not None
        matchable_aids = qreq_.daids
        _gtaids = ibs.get_annot_groundtruth(cm.qaid, daid_list=matchable_aids)

        if viz_name_score:
            # Only look at the groundtruth if a name isnt in the top list
            _gtnids = ibs.get_annot_name_rowids(_gtaids)
            top_nids = ibs.get_annot_name_rowids(top_aids)
            _valids = ~np.in1d(_gtnids, top_nids)
            _gtaids = ut.compress(_gtaids, _valids)

        # No need to display highly ranked groundtruth. It will already show up
        _gtaids = np.setdiff1d(_gtaids, top_aids)
        # Sort missed grountruth by score
        _gtscores = cm.get_annot_scores(_gtaids)
        _gtaids = ut.sortedby(_gtaids, _gtscores, reverse=True)
        if viz_name_score:
            if len(_gtaids) > 1:
                _gtaids = _gtaids[0:1]
        else:
            if len(_gtaids) > 3:
                # Hack to not show too many unmatched groundtruths
                # _isexmp = ibs.get_annot_exemplar_flags(_gtaids)
                _gtaids = _gtaids[0:3]
        showgt_aids = _gtaids

    if show_gf:
        # Show only one top-scoring groundfalse example
        top_nids = ibs.get_annot_name_rowids(top_aids)
        is_groundfalse = top_nids != ibs.get_annot_name_rowids(cm.qaid)
        gf_idxs = np.nonzero(is_groundfalse)[0]
        if len(gf_idxs) > 0:
            best_gf_idx = gf_idxs[0]
            isvalid = ~is_groundfalse
            isvalid[best_gf_idx] = True
            # Filter so there is only one groundfalse
            top_aids = top_aids.compress(isvalid)
        # else:
        #     if len(top_aids) == 1:
        #         top_aids = top_aids.tolist()
        #         top_aids.append(None)

        # else:
        #     # seems like there were no results. Must be bad feature detections
        #     # maybe too much spatial verification
        #     top_aids = []

        # if len(showgt_aids) != 0:
        #     # Hack to just include gtaids in normal list
        #     top_aids = np.append(top_aids, showgt_aids)
        #     showgt_aids = []

    if viz_name_score:
        # Make sure that there is only one of each name in the list
        top_nids = ibs.get_annot_name_rowids(top_aids)
        top_aids = ut.compress(top_aids, ut.flag_unique_items(top_nids))

    return show_qres(
        ibs,
        cm,
        gt_aids=showgt_aids,
        top_aids=top_aids,
        figtitle=figtitle,
        show_query=show_query,
        qreq_=qreq_,
        failed_to_match=failed_to_match,
        **kwargs,
    )


def show_qres(ibs, cm, qreq_=None, **kwargs):
    r"""
    Display Query Result Logic
    Defaults to: query chip, groundtruth matches, and top matches

    Args:
        ibs (wbia.IBEISController): wbia controller object
        cm (wbia.ChipMatch): object of feature correspondences and scores
        qreq_ (wbia.QueryRequest):  query request object with hyper-parameters(default = None)

    Kwargs:
        annot_mode, figtitle, make_figtitle, aug, top_aids, all_kpts,
        show_query, in_image, sidebyside, name_scoring, max_nCols,
        failed_to_match, fnum
        in_image (bool) show result  in image view if True else chip view
        annot_mode (int):
            if annot_mode == 0, then draw lines and ellipse
            elif annot_mode == 1, then dont draw lines or ellipse
            elif annot_mode == 2, then draw only lines
            elif annot_mode == 3, draw heatmask only
        See: viz_matches.show_name_matches, viz_helpers.get_query_text

    Returns:
        mpl.Figure: fig

    CommandLine:
        ./main.py --query 1 -y --db PZ_MTEST --noshow-qres
        python -m wbia.viz.viz_qres show_qres --show
        python -m wbia.viz.viz_qres show_qres --show --top-aids=10 --db=PZ_MTEST \
                --sidebyside --annot_mode=0 --notitle --no-viz_name_score \
                --qaids=5 --max_nCols=2 --adjust=.01,.01,.01

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.viz.viz_qres import *  # NOQA
        >>> import wbia
        >>> cm, qreq_ = wbia.testdata_cm()
        >>> kwargs = dict(
        >>>     top_aids=ut.get_argval('--top-aids', type_=int, default=3),
        >>>     sidebyside=not ut.get_argflag('--no-sidebyside'),
        >>>     annot_mode=ut.get_argval('--annot_mode', type_=int, default=1),
        >>>     viz_name_score=not ut.get_argflag('--no-viz_name_score'),
        >>>     simplemode=ut.get_argflag('--simplemode'),
        >>>     max_nCols=ut.get_argval('--max_nCols', type_=int, default=None)
        >>> )
        >>> ibs = qreq_.ibs
        >>> fig = show_qres(ibs, cm, show_query=False, qreq_=qreq_, **kwargs)
        >>> ut.show_if_requested()
    """
    # ut.print_dict(kwargs)
    annot_mode = kwargs.get('annot_mode', 1) % 4  # this is toggled
    figtitle = kwargs.get('figtitle', '')
    aug = kwargs.get('aug', '')
    top_aids = kwargs.get('top_aids', DEFAULT_NTOP)
    gt_aids = kwargs.get('gt_aids', [])
    all_kpts = kwargs.get('all_kpts', False)
    show_query = kwargs.get('show_query', False)
    in_image = kwargs.get('in_image', False)
    sidebyside = kwargs.get('sidebyside', True)
    simplemode = kwargs.get('simplemode', False)
    colorbar_ = kwargs.get('colorbar_', False)
    # name_scoring   = kwargs.get('name_scoring', False)
    viz_name_score = kwargs.get('viz_name_score', qreq_ is not None)
    max_nCols = kwargs.get('max_nCols', None)
    failed_to_match = kwargs.get('failed_to_match', False)

    fnum = pt.ensure_fnum(kwargs.get('fnum', None))

    if ut.VERBOSE and ut.NOT_QUIET:
        logger.info(
            'query_info = '
            + ut.repr2(
                ibs.get_annot_info(
                    cm.qaid,
                    default=True,
                    gname=False,
                    name=False,
                    notes=False,
                    exemplar=False,
                ),
                nl=4,
            )
        )
        logger.info(
            'top_aids_info = '
            + ut.repr2(
                ibs.get_annot_info(
                    top_aids,
                    default=True,
                    gname=False,
                    name=False,
                    notes=False,
                    exemplar=False,
                    reference_aid=cm.qaid,
                ),
                nl=4,
            )
        )

    fig = pt.figure(fnum=fnum, docla=True, doclf=True)

    if isinstance(top_aids, int):
        top_aids = cm.get_top_aids(top_aids)

    if failed_to_match:
        # HACK to visually indicate failure to match in analysis
        show_query = True
        top_aids = [None] + top_aids

    nTop = len(top_aids)

    if max_nCols is None:
        max_nCols = 5
        if nTop in [6, 7]:
            max_nCols = 3
        if nTop in [8]:
            max_nCols = 4

    try:
        assert (
            len(list(set(top_aids).intersection(set(gt_aids)))) == 0
        ), 'gts should be missed.  not in top'
    except AssertionError as ex:
        ut.printex(ex, keys=['top_aids', 'gt_aids'])
        raise

    if ut.DEBUG2:
        logger.info(cm.get_inspect_str())

    # --------------------------------------------------
    # Get grid / cell information to build subplot grid
    # --------------------------------------------------
    # Show query or not
    nQuerySubplts = 1 if show_query else 0
    # The top row is given slots for ground truths and querys
    # all aids in gt_aids should not be in top aids
    nGtSubplts = nQuerySubplts + (0 if gt_aids is None else len(gt_aids))
    # The bottom rows are for the top results
    nTopNSubplts = nTop
    nTopNCols = min(max_nCols, nTopNSubplts)
    nGTCols = min(max_nCols, nGtSubplts)
    nGTCols = max(nGTCols, nTopNCols)
    nTopNCols = nGTCols
    # Get number of rows to show groundtruth
    nGtRows = 0 if nGTCols == 0 else int(np.ceil(nGtSubplts / nGTCols))
    # Get number of rows to show results
    nTopNRows = 0 if nTopNCols == 0 else int(np.ceil(nTopNSubplts / nTopNCols))
    nGtCells = nGtRows * nGTCols
    # Total number of rows
    nRows = nTopNRows + nGtRows

    DEBUG_SHOW_QRES = 0

    if DEBUG_SHOW_QRES:
        allgt_aids = ibs.get_annot_groundtruth(cm.qaid)
        nSelGt = len(gt_aids)
        nAllGt = len(allgt_aids)
        logger.info('[show_qres]========================')
        logger.info('[show_qres]----------------')
        logger.info('[show_qres] * annot_mode=%r' % (annot_mode,))
        logger.info('[show_qres] #nTop=%r #missed_gts=%r/%r' % (nTop, nSelGt, nAllGt))
        logger.info('[show_qres] * -----')
        logger.info('[show_qres] * nRows=%r' % (nRows,))
        logger.info('[show_qres] * nGtSubplts=%r' % (nGtSubplts,))
        logger.info('[show_qres] * nTopNSubplts=%r' % (nTopNSubplts,))
        logger.info('[show_qres] * nQuerySubplts=%r' % (nQuerySubplts,))
        logger.info('[show_qres] * -----')
        logger.info('[show_qres] * nGTCols=%r' % (nGTCols,))
        logger.info('[show_qres] * -----')
        logger.info('[show_qres] * fnum=%r' % (fnum,))
        logger.info('[show_qres] * figtitle=%r' % (figtitle,))
        logger.info('[show_qres] * max_nCols=%r' % (max_nCols,))
        logger.info('[show_qres] * show_query=%r' % (show_query,))
        logger.info('[show_qres] * kwargs=%s' % (ut.repr2(kwargs),))

    # HACK:
    _color_list = pt.distinct_colors(nTop)
    aid2_color = {aid: _color_list[ox] for ox, aid in enumerate(top_aids)}
    ranked_aids = cm.get_top_aids()

    # Helpers
    def _show_query_fn(plotx_shift, rowcols):
        """ helper for show_qres """
        plotx = plotx_shift + 1
        pnum = (rowcols[0], rowcols[1], plotx)
        # logger.info('[viz] Plotting Query: pnum=%r' % (pnum,))
        _kwshow = dict(draw_kpts=annot_mode)
        _kwshow.update(kwargs)
        _kwshow['prefix'] = 'q'
        _kwshow['pnum'] = pnum
        _kwshow['aid2_color'] = aid2_color
        _kwshow['draw_ell'] = annot_mode >= 1
        viz_chip.show_chip(ibs, cm.qaid, annote=False, qreq_=qreq_, **_kwshow)

    def _plot_matches_aids(aid_list, plotx_shift, rowcols):
        """ helper for show_qres to draw many aids """
        _kwshow = dict(
            draw_ell=annot_mode,
            draw_pts=False,
            draw_lines=annot_mode,
            ell_alpha=0.5,
            all_kpts=all_kpts,
        )
        _kwshow.update(kwargs)
        _kwshow['fnum'] = fnum
        _kwshow['in_image'] = in_image
        _kwshow['colorbar_'] = colorbar_
        if sidebyside:
            # Draw each match side by side the query
            _kwshow['draw_ell'] = annot_mode in {1}
            _kwshow['draw_lines'] = annot_mode in {1, 2}
            _kwshow['heatmask'] = annot_mode in {3}
        else:
            # logger.info('annot_mode = %r' % (annot_mode,))
            _kwshow['draw_ell'] = annot_mode == 1
            # _kwshow['draw_pts'] = annot_mode >= 1
            # _kwshow['draw_lines'] = False
            _kwshow['show_query'] = False

        def _show_matches_fn(aid, orank, pnum):
            """ Helper function for drawing matches to one aid """
            aug = 'rank=%r\n' % orank
            _kwshow['pnum'] = pnum
            _kwshow['title_aug'] = aug
            # draw_ell = annot_mode == 1
            # draw_lines = annot_mode >= 1
            # If we already are showing the query dont show it here
            if sidebyside:
                # Draw each match side by side the query
                if viz_name_score:
                    cm.show_single_namematch(qreq_, ibs.get_annot_nids(aid), **_kwshow)
                else:
                    if simplemode:
                        _kwshow['draw_border'] = False
                        _kwshow['draw_lbl'] = False
                        _kwshow['notitle'] = True
                        _kwshow['vert'] = False
                        _kwshow['modifysize'] = True
                    cm.show_single_annotmatch(qreq_, aid, **_kwshow)
                    # viz_matches.show_matches(ibs, cm, aid, qreq_=qreq_, **_kwshow)
            else:
                # Draw each match by themselves
                data_config2_ = None if qreq_ is None else qreq_.extern_data_config2
                # _kwshow['draw_border'] = kwargs.get('draw_border', True)
                # _kwshow['notitle'] = ut.get_argflag(('--no-title', '--notitle'))
                viz_chip.show_chip(
                    ibs,
                    aid,
                    annote=False,
                    notitle=True,
                    data_config2_=data_config2_,
                    **_kwshow,
                )

        if DEBUG_SHOW_QRES:
            logger.info('[show_qres()] Plotting Chips %s:' % vh.get_aidstrs(aid_list))
        if aid_list is None:
            return
        # Do lazy load before show
        # data_config2_ = None if qreq_ is None else qreq_.extern_data_config2

        # tblhack = getattr(qreq_, 'tablename', None)
        # HACK FOR HUMPBACKS
        # (Also in viz_matches)
        # if tblhack == 'vsone' or (qreq_ is not None and not qreq_._isnewreq):
        #     # precompute
        #     pass
        #     #ibs.get_annot_chips(aid_list, config2_=data_config2_, ensure=True)
        #     #ibs.get_annot_kpts(aid_list, config2_=data_config2_, ensure=True)

        for ox, aid in enumerate(aid_list):
            plotx = ox + plotx_shift + 1
            pnum = (rowcols[0], rowcols[1], plotx)
            oranks = np.where(ranked_aids == aid)[0]
            # This pair has no matches between them.
            if len(oranks) == 0:
                orank = -1
                if aid is None:
                    pt.imshow_null(
                        'Failed to find matches\nfor qaid=%r' % (cm.qaid),
                        fnum=fnum,
                        pnum=pnum,
                        fontsize=18,
                    )
                else:
                    _show_matches_fn(aid, orank, pnum)
                # if DEBUG_SHOW_QRES:
                #    logger.info('skipping pnum=%r' % (pnum,))
                continue
            if DEBUG_SHOW_QRES:
                logger.info('pnum=%r' % (pnum,))
            orank = oranks[0] + 1
            _show_matches_fn(aid, orank, pnum)

    shift_topN = nGtCells

    if nGtSubplts == 1:
        nGTCols = 1

    if nRows == 0:
        pt.imshow_null('[viz_qres] No matches. nRows=0', fnum=fnum)
    else:
        fig = pt.figure(fnum=fnum, pnum=(nRows, nGTCols, 1), docla=True, doclf=True)
        pt.plt.subplot(nRows, nGTCols, 1)
        # Plot Query
        if show_query:
            _show_query_fn(0, (nRows, nGTCols))
        # Plot Ground Truth (if given)
        _plot_matches_aids(gt_aids, nQuerySubplts, (nRows, nGTCols))
        # Plot Results
        _plot_matches_aids(top_aids, shift_topN, (nRows, nTopNCols))
        figtitle += aug
    if failed_to_match:
        figtitle += '\n No matches found'

    incanvas = kwargs.get('with_figtitle', not vh.NO_LBL_OVERRIDE)
    pt.set_figtitle(figtitle, incanvas=incanvas)

    # Result Interaction
    return fig



# -*- coding: utf-8 -*-
import logging
import utool as ut
from six.moves import range
from wbia.plottool import draw_func2 as df2
from wbia.plottool.viz_featrow import draw_feat_row
from wbia.viz import viz_helpers as vh
import wbia.plottool as pt  # NOQA
import six  # NOQA

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def get_annotfeat_nn_index(ibs, qaid, qfx, qreq_=None):
    # raise NotImplementedError('this doesnt work anymore. Need to submit mc4 query with metadata on and then reextract the required params')
    # from . import match_chips3 as mc3
    # ibs._init_query_requestor()
    if qreq_ is None:
        daid_list = ibs.get_valid_aids()
        qreq_ = ibs.new_query_request([qaid], daid_list)
    qreq_.load_indexer()  # TODO: ensure lazy

    # if isinstance(qfx, six.string_types):
    special = qfx == 'special'
    if special:
        qfx2_vecs = ibs.get_annot_vecs(qaid)
    else:
        qfx = int(qfx)
        qfx2_vecs = ibs.get_annot_vecs(qaid)[qfx : (qfx + 1)]
    K = qreq_.qparams.K
    Knorm = qreq_.qparams.Knorm
    if ut.VERBOSE:
        logger.info('Knorm = %r' % (Knorm,))
    qfx2_idx, qfx2_dist = qreq_.indexer.knn(qfx2_vecs, 10)

    if special:
        import numpy as np

        # Find a query feature with "good" results
        qfx2_daid = qreq_.indexer.get_nn_aids(qfx2_idx)
        qfx2_dnid = ibs.get_annot_nids(qfx2_daid)
        nid = ibs.get_annot_nids(qaid)

        # slice_ = slice(None)
        slice_ = slice(0, K + Knorm)
        flags = qfx2_dnid.T[slice_].T == nid
        flags = np.logical_and(flags, qfx2_daid[:, slice_] != qaid)
        flags_first = flags[:, 0:K]
        flags_last = flags[:, K:]
        num_gt_matches = flags_first.sum(axis=1) - flags_last.sum(axis=1)
        logger.info('num_gt_matches = %r' % (num_gt_matches,))
        logger.info(num_gt_matches.max())
        has_good_num = num_gt_matches >= num_gt_matches.max() - 1
        candidate_qfxs = np.where(has_good_num)[0]

        cand_nids = qfx2_dnid[candidate_qfxs].T[slice_].T
        cand_flags = cand_nids == nid
        cand_dist = qfx2_dist[candidate_qfxs].T[slice_].T
        cand_dist_gt = cand_dist * cand_flags
        cand_dist_gf = cand_dist * ~cand_flags
        cand_score = cand_dist_gt.sum(axis=1) - cand_dist_gf.sum(axis=1)
        top_candxs = cand_score.argsort()
        logger.info('cand_nids = %r' % (cand_nids,))
        logger.info('top_candxs = %r' % (top_candxs,))

        cand_idx = top_candxs[1]
        # cand_idx = ut.take_percentile(top_candxs, .1)[-1]
        qfx = candidate_qfxs[cand_idx]
        logger.info('qfx = %r' % (qfx,))
        qfx2_dist = qfx2_dist[qfx : (qfx + 1)]
        qfx2_idx = qfx2_idx[qfx : (qfx + 1)]

    qfx2_daid = qreq_.indexer.get_nn_aids(qfx2_idx)
    qfx2_dfx = qreq_.indexer.get_nn_featxs(qfx2_idx)
    return qfx, qfx2_daid, qfx2_dfx, qfx2_dist, K, Knorm


def show_top_featmatches(qreq_, cm_list):
    """
    Args:
        qreq_ (wbia.QueryRequest):  query request object with hyper-parameters
        cm_list (list):

    SeeAlso:
        python -m wbia --tf TestResult.draw_feat_scoresep --show --db PZ_MTEST -t best:lnbnn_on=True,lnbnn_normalizer=normlnbnn-test -a default --sephack

        python -m wbia --tf TestResult.draw_feat_scoresep --show --db PZ_Master1 -t best:lnbnn_on=True -a timectrl --sephack
        python -m wbia --tf TestResult.draw_feat_scoresep --show --db PZ_MTEST -t best:lnbnn_on=True -a default:size=30 --sephack
        python -m wbia --tf TestResult.draw_feat_scoresep --show --db PZ_MTEST -t best:K=1,Knorm=5,lnbnn_on=True -a default:size=30 --sephack
        python -m wbia --tf TestResult.draw_feat_scoresep --show --db PZ_MTEST -t best:K=1,Knorm=3,lnbnn_on=True -a default --sephack


    CommandLine:
        python -m wbia.viz.viz_nearest_descriptors --exec-show_top_featmatches --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.viz.viz_nearest_descriptors import *  # NOQA
        >>> import wbia
        >>> cm_list, qreq_ = wbia.testdata_cmlist(defaultdb='PZ_MTEST',
        >>>                                        a=['default:has_none=mother,size=30'])
        >>> show_top_featmatches(qreq_, cm_list)
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> ut.show_if_requested()
    """
    # for cm in cm_list:
    #     cm.score_annot_csum(qreq_)
    import numpy as np
    import vtool as vt
    from functools import partial

    # Stack chipmatches
    ibs = qreq_.ibs
    infos = [cm.get_flat_fm_info() for cm in cm_list]
    flat_metadata = dict(
        [(k, np.concatenate(v)) for k, v in ut.dict_stack2(infos).items()]
    )
    fsv_flat = flat_metadata['fsv']
    flat_metadata['fs'] = fsv_flat.prod(axis=1)
    aids1 = flat_metadata['aid1'][:, None]
    aids2 = flat_metadata['aid2'][:, None]
    flat_metadata['aid_pairs'] = np.concatenate([aids1, aids2], axis=1)

    # Take sample of metadata
    sortx = flat_metadata['fs'].argsort()[::-1]
    num = len(cm_list) * 3
    # num = 10
    taker = partial(np.take, indices=sortx[:num], axis=0)
    flat_metadata_top = ut.map_dict_vals(taker, flat_metadata)
    aid1s, aid2s, fms = ut.dict_take(flat_metadata_top, ['aid1', 'aid2', 'fm'])

    annots = {}
    aids = np.unique(np.hstack((aid1s, aid2s)))
    annots = {aid: ibs.get_annot_lazy_dict(aid, config2_=qreq_.qparams) for aid in aids}

    label_lists = (
        ibs.get_match_truths(aid1s, aid2s) == ibs.const.EVIDENCE_DECISION.POSITIVE
    )
    patch_size = 64

    def extract_patches(annots, aid, fxs):
        """ custom_func(lazydict, key, subkeys) for multigroup_lookup """
        annot = annots[aid]
        kpts = annot['kpts']
        rchip = annot['rchip']
        kpts_m = kpts.take(fxs, axis=0)
        warped_patches, warped_subkpts = vt.get_warped_patches(
            rchip, kpts_m, patch_size=patch_size
        )
        return warped_patches

    data_lists = vt.multigroup_lookup(annots, [aid1s, aid2s], fms.T, extract_patches)

    import wbia.plottool as pt  # NOQA

    pt.ensureqt()
    import wbia_cnn

    inter = wbia_cnn.draw_results.interact_patches(
        label_lists,
        data_lists,
        flat_metadata_top,
        chunck_sizes=(2, 4),
        ibs=ibs,
        hack_one_per_aid=False,
        sortby='fs',
        qreq_=qreq_,
    )
    inter.show()


# @ut.indent_func('[show_neardesc]')
def show_nearest_descriptors(ibs, qaid, qfx, fnum=None, stride=5, qreq_=None, **kwargs):
    r"""
    Args:
        ibs (wbia.IBEISController): image analysis api
        qaid (int):  query annotation id
        qfx (int): query feature index
        fnum (int):  figure number
        stride (int):
        consecutive_distance_compare (bool):

    CommandLine:
        # Find a good match to inspect
        python -m wbia.viz.interact.interact_matches --test-testdata_match_interact --show --db PZ_MTEST --qaid 3

        # Now inspect it
        python -m wbia.viz.viz_nearest_descriptors --test-show_nearest_descriptors --show --db PZ_MTEST --qaid 3 --qfx 879
        python -m wbia.viz.viz_nearest_descriptors --test-show_nearest_descriptors --show
        python -m wbia.viz.viz_nearest_descriptors --test-show_nearest_descriptors --db PZ_MTEST --qaid 3 --qfx 879 --diskshow --save foo.png --dpi=256

    SeeAlso:
        plottool.viz_featrow
        ~/code/plottool/plottool/viz_featrow.py

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.viz.viz_nearest_descriptors import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> if True:
        >>>     import matplotlib as mpl
        >>>     from wbia.scripts.thesis import TMP_RC
        >>>     mpl.rcParams.update(TMP_RC)
        >>> qreq_ = wbia.testdata_qreq_()
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> qaid = qreq_.qaids[0]
        >>> qfx = ut.get_argval('--qfx', type_=None, default=879)
        >>> fnum = None
        >>> stride = 5
        >>> # execute function
        >>> skip = False
        >>> result = show_nearest_descriptors(ibs, qaid, qfx, fnum, stride,
        >>>                                   draw_chip=True,
        >>>                                   draw_warped=True,
        >>>                                   draw_unwarped=False,
        >>>                                   draw_desc=False, qreq_=qreq_)
        >>> # verify results
        >>> print(result)
        >>> pt.show_if_requested()
    """
    import wbia.plottool as pt  # NOQA

    consecutive_distance_compare = True
    draw_chip = kwargs.get('draw_chip', False)
    draw_desc = kwargs.get('draw_desc', True)
    draw_warped = kwargs.get('draw_warped', True)
    draw_unwarped = kwargs.get('draw_unwarped', True)
    # skip = kwargs.get('skip', True)
    # Plots the nearest neighbors of a given feature (qaid, qfx)
    if fnum is None:
        fnum = df2.next_fnum()
    try:
        # Flann NN query
        (qfx, qfx2_daid, qfx2_dfx, qfx2_dist, K, Knorm) = get_annotfeat_nn_index(
            ibs, qaid, qfx, qreq_=qreq_
        )

        # Adds metadata to a feature match
        def get_extract_tuple(aid, fx, k=-1):
            rchip = ibs.get_annot_chips(aid)
            kp = ibs.get_annot_kpts(aid)[fx]
            sift = ibs.get_annot_vecs(aid)[fx]
            if not ut.get_argflag('--texknormplot'):
                aidstr = vh.get_aidstrs(aid)
                nidstr = vh.get_nidstrs(ibs.get_annot_nids(aid))
                id_str = ' ' + aidstr + ' ' + nidstr + ' fx=%r' % (fx,)
            else:
                id_str = nidstr = aidstr = ''
            info = ''
            if k == -1:
                if pt.is_texmode():
                    info = '\\vspace{1cm}'
                    info += 'Query $\\mathbf{d}_i$'
                    info += '\n\\_'
                    info += '\n\\_'
                else:
                    if len(id_str) > '':
                        info = 'Query: %s' % (id_str,)
                    else:
                        info = 'Query'
                type_ = 'Query'
            elif k < K:
                type_ = 'Match'
                if ut.get_argflag('--texknormplot') and pt.is_texmode():
                    # info = 'Match:\n$k=%r$, $\\frac{||\\mathbf{d}_i - \\mathbf{d}_j||}{Z}=%.3f$' % (k, qfx2_dist[0, k])
                    info = '\\vspace{1cm}'
                    info += 'Match: $\\mathbf{d}_{j_%r}$\n$\\textrm{dist}=%.3f$' % (
                        k,
                        qfx2_dist[0, k],
                    )
                    # info += '\n$s_{\\tt{LNBNN}}=%.3f$' % (qfx2_dist[0, K + Knorm - 1] - qfx2_dist[0, k])
                    info += '\n$s=%.3f$' % (qfx2_dist[0, K + Knorm - 1] - qfx2_dist[0, k])
                else:
                    info = 'Match:%s\nk=%r, dist=%.3f' % (id_str, k, qfx2_dist[0, k])
                    info += '\nLNBNN=%.3f' % (
                        qfx2_dist[0, K + Knorm - 1] - qfx2_dist[0, k]
                    )
            elif k < Knorm + K:
                type_ = 'Norm'
                if ut.get_argflag('--texknormplot') and pt.is_texmode():
                    # info = 'Norm: $j_%r$\ndist=%.3f' % (id_str, k, qfx2_dist[0, k])
                    info = '\\vspace{1cm}'
                    info += 'Norm: $j_%r$\n$\\textrm{dist}=%.3f$' % (k, qfx2_dist[0, k])
                    info += '\n\\_'
                else:
                    info = 'Norm: %s\n$k=%r$, dist=$%.3f$' % (id_str, k, qfx2_dist[0, k])
            else:
                raise Exception('[viz] problem k=%r')
            return (rchip, kp, sift, fx, aid, info, type_)

        extracted_list = []
        # Remember the query sift feature
        extracted_list.append(get_extract_tuple(qaid, qfx, -1))
        origsift = extracted_list[0][2]
        skipped = 0
        for k in range(K + Knorm):
            # if qfx2_daid[0, k] == qaid and qfx2_dfx[0, k] == qfx:
            if qfx2_daid[0, k] == qaid:
                skipped += 1
                continue
            tup = get_extract_tuple(qfx2_daid[0, k], qfx2_dfx[0, k], k)
            extracted_list.append(tup)
        # Draw the _select_ith_match plot
        nRows = len(extracted_list)
        if stride is None:
            stride = nRows
        # Draw selected feature matches
        prevsift = None
        px = 0  # plot offset
        px_shift = 0  # plot stride shift
        nExtracted = len(extracted_list)
        featrow_kw = dict(
            draw_chip=draw_chip,
            draw_desc=draw_desc,
            draw_warped=draw_warped,
            draw_unwarped=draw_unwarped,
        )
        if ut.get_argflag('--texknormplot'):
            featrow_kw['ell_color'] = pt.ORANGE
            featrow_kw['ell_linewidth'] = 1
            featrow_kw['arm1_lw'] = 0.5
            featrow_kw['stroke'] = 0
            pass
        for listx, tup in enumerate(extracted_list):
            (rchip, kp, sift, fx, aid, info, type_) = tup
            if listx % stride == 0:
                # Create a temporary nRows and fnum in case we are splitting
                # up nearest neighbors into separate figures with stride
                _fnum = fnum + listx
                _nRows = min(nExtracted - listx, stride)
                px_shift = px
                df2.figure(fnum=_fnum, docla=True, doclf=True)
            px_ = px - px_shift
            px = draw_feat_row(
                rchip,
                fx,
                kp,
                sift,
                _fnum,
                _nRows,
                px=px_,
                prevsift=prevsift,
                origsift=origsift,
                aid=aid,
                info=info,
                type_=type_,
                **featrow_kw,
            )

            px += px_shift
            if prevsift is None or consecutive_distance_compare:
                prevsift = sift

        # df2.adjust_subplots(hspace=.85, wspace=0, top=.95, bottom=.087, left=.05, right=.95)

    except Exception as ex:
        logger.info('[viz] Error in show nearest descriptors')
        logger.info(ex)
        raise



# -*- coding: utf-8 -*-
### __init__.py ###
# flake8: noqa
import logging
from wbia.expt import experiment_configs
from wbia.expt import experiment_drawing
from wbia.expt import test_result
from wbia.expt import harness
from wbia.expt import experiment_helpers
from wbia.expt import experiment_printres
import utool as ut

print, rrr, profile = ut.inject2(__name__, '[wbia.expt]')
logger = logging.getLogger('wbia')


def reassign_submodule_attributes(verbose=True):
    """
    why reloading all the modules doesnt do this I don't know
    """
    import sys

    if verbose and '--quiet' not in sys.argv:
        logger.info('dev reimport')
    # Self import
    import wbia.expt

    # Implicit reassignment.
    seen_ = set([])
    for tup in IMPORT_TUPLES:
        if len(tup) > 2 and tup[2]:
            continue  # dont import package names
        submodname, fromimports = tup[0:2]
        submod = getattr(wbia.expt, submodname)
        for attr in dir(submod):
            if attr.startswith('_'):
                continue
            if attr in seen_:
                # This just holds off bad behavior
                # but it does mimic normal util_import behavior
                # which is good
                continue
            seen_.add(attr)
            setattr(wbia.expt, attr, getattr(submod, attr))


def reload_subs(verbose=True):
    """ Reloads wbia.expt and submodules """
    rrr(verbose=verbose)

    def fbrrr(*args, **kwargs):
        """ fallback reload """
        pass

    getattr(experiment_configs, 'rrr', fbrrr)(verbose=verbose)
    getattr(harness, 'rrr', fbrrr)(verbose=verbose)
    getattr(experiment_helpers, 'rrr', fbrrr)(verbose=verbose)
    getattr(experiment_printres, 'rrr', fbrrr)(verbose=verbose)
    getattr(results_all, 'rrr', fbrrr)(verbose=verbose)
    getattr(results_analyzer, 'rrr', fbrrr)(verbose=verbose)
    getattr(results_organizer, 'rrr', fbrrr)(verbose=verbose)
    rrr(verbose=verbose)
    try:
        # hackish way of propogating up the new reloaded submodule attributes
        reassign_submodule_attributes(verbose=verbose)
    except Exception as ex:
        logger.info(ex)


rrrr = reload_subs

IMPORT_TUPLES = [
    ('experiment_configs', None),
    ('harness', None),
    ('experiment_helpers', None),
    ('experiment_printres', None),
    ('results_all', None),
    ('results_analyzer', None),
    ('results_organizer', None),
]
"""
Regen Command:
    cd /home/joncrall/code/wbia/wbia.expt
    makeinit.py
"""
# autogenerated __init__.py for: '/home/joncrall/code/wbia/wbia.expt'



# -*- coding: utf-8 -*-
"""
IBEIS: main package init

TODO: LAZY IMPORTS?
    http://code.activestate.com/recipes/473888-lazy-module-imports/
"""
# flake8: noqa

try:
    from wbia._version import __version__
except ImportError:
    __version__ = '0.0.0'

try:
    import utool as ut
    from wbia import dtool
except ImportError as ex:
    print('[wbia !!!] ERROR: Unable to load all core utility modules.')
    print('[wbia !!!] Perhaps try super_setup.py pull')
    raise

ut.noinject(__name__, '[wbia.__init__]')
if ut.VERBOSE:
    print('[wbia] importing wbia __init__')


if ut.is_developer():
    standard_visualization_functions = [
        'show_image',
        'show_chip',
        'show_chipmatch',
        'show_chipmatches',
        'show_vocabulary',
        #'show_vocabulary',
    ]

# If we dont initialize plottool before <something>
# then it causes a crash in windows. Its so freaking weird.
# something is not guitool, wbia.viz
# has to be before control, can be after constants, params, and entry_points
# import wbia.plottool


ENABLE_WILDBOOK_SIGNAL = False


try:
    from wbia import constants
    from wbia import constants as const
    from wbia import params
    from wbia import entry_points
    from wbia import other
    from wbia.init import sysres

    # entry_points._preload()

    from wbia import control
    from wbia import dbio

    # from wbia import web

    from wbia.init import sysres
    from wbia.entry_points import (
        main,
        _preload,
        _init_numpy,
        main_loop,
        opendb,
        opendb_in_background,
        opendb_bg_web,
    )
    from wbia.control.IBEISControl import IBEISController
    from wbia.algo.hots.query_request import QueryRequest
    from wbia.algo.hots.chip_match import ChipMatch, AnnotMatch
    from wbia.algo.graph.core import AnnotInference
    from wbia.init.sysres import (
        get_workdir,
        set_workdir,
        ensure_pz_mtest,
        ensure_nauts,
        ensure_wilddogs,
        list_dbs,
    )
    from wbia.init import main_helpers

    from wbia import algo

    from wbia import expt
    from wbia import templates
    from wbia.templates import generate_notebook
    from wbia.control.controller_inject import register_preprocs
    from wbia import core_annots
    from wbia import core_images

    try:
        from wbia.scripts import postdoc
    except ImportError:
        pass
except Exception as ex:
    ut.printex(ex, 'Error when importing wbia', tb=True)
    raise


def import_subs():
    # Weird / Fancy loading.
    # I want to make this simpler
    from wbia import algo
    from wbia import viz
    from wbia import web
    from wbia import gui
    from wbia import templates


def run_experiment(
    e='print',
    db='PZ_MTEST',
    dbdir=None,
    a=['unctrl'],
    t=['default'],
    initial_aids=None,
    qaid_override=None,
    daid_override=None,
    lazy=False,
    **kwargs,
):
    """
    Convience function

    CommandLine:
        wbia -e print

    Args:
        e (str): (default = 'print')
        db (str): (default = 'PZ_MTEST')
        a (list): (default = ['unctrl'])
        t (list): (default = ['default'])
        qaid_override (None): (default = None)
        lazy (bool): (default = False)

    Returns:
        function: func -  live python function

    CommandLine:
        python -m wbia.__init__ --exec-run_experiment --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia import *  # NOQA
        >>> e = 'rank_cmc'
        >>> db = 'testdb1'
        >>> a = ['default:species=primary']
        >>> t = ['default']
        >>> initial_aids = [2, 3, 4, 7, 9, 10, 11]
        >>> qaid_override = [1, 9, 10, 11, 2, 3]
        >>> testres = run_experiment(e, db, a, t, qaid_override=qaid_override,
        >>>                          initial_aids=initial_aids)
        >>> result = ('testres = %s' % (str(testres),))
        >>> print(result)
        >>> ut.quit_if_noshow()
        >>> testres.draw_func()
        >>> ut.show_if_requested()
    """
    import functools

    def find_expt_func(e):
        import utool as ut
        import wbia.dev

        for tup in wbia.dev.REGISTERED_DOCTEST_EXPERIMENTS:
            modname, funcname = tup[:2]
            aliases = tup[2] if len(tup) == 3 else []
            if e == funcname or e in aliases:
                module = ut.import_modname(modname)
                func = module.__dict__[funcname]
                return func
        # hack in --tf magic
        func = ut.find_testfunc('wbia', funcname)[0]
        return func

    def build_commandline(e=e, **kwargs):
        # Equivalent command line version of this func
        import wbia.dev

        valid_e_flags = ut.flatten(
            [
                [tup[1]] if len(tup) == 2 else [tup[1]] + tup[2]
                for tup in wbia.dev.REGISTERED_DOCTEST_EXPERIMENTS
            ]
        )
        if e in valid_e_flags:
            epref = '-e'
        else:
            # hack to use tf
            epref = '--tf'

        if dbdir is not None:
            db_flag = '--dbdir'
            db_value = dbdir
        else:
            db_flag = '--db'
            db_value = db
        command_parts = [
            'wbia',
            epref,
            e,
            db_flag,
            db_value,
            '-a',
            ' '.join(a).replace('(', r'\(').replace(')', r'\)'),
            '-t',
            ' '.join(t),
        ]
        if qaid_override is not None:
            command_parts.extend(['--qaid=' + ','.join(map(str, qaid_override))])
        if daid_override is not None:
            command_parts.extend(['--daid-override=' + ','.join(map(str, daid_override))])
        if 'disttype' in kwargs:
            command_parts.extend(['--disttype=', ','.join(map(str, kwargs['disttype']))])

        # hack parse out important args that were on command line
        if 'f' in kwargs:
            command_parts.extend(['-f', ' '.join(kwargs['f'])])
        if 'test_cfgx_slice' in kwargs:
            # very hacky, much more than checking for f
            slice_ = kwargs['test_cfgx_slice']
            slice_attrs = [
                getattr(slice_, attr, '') for attr in ['start', 'stop', 'step']
            ]
            slice_attrs = ut.replace_nones(slice_attrs, '')
            slicestr = ':'.join(map(str, slice_attrs))
            command_parts.extend(['--test_cfgx_slice', slicestr])

        command_parts.extend(['--show'])

        command_line_str = ' '.join(command_parts)
        # Warning, not always equivalent
        print('Equivalent Command Line:')
        print(command_line_str)
        return command_line_str

    command_line_str = build_commandline(**kwargs)

    def draw_cases(testres, **kwargs):
        e_ = 'draw_cases'
        func = find_expt_func(e_)
        ibs = testres.ibs
        build_commandline(e=e_, **kwargs)
        lazy_func = functools.partial(func, ibs, testres, show_in_notebook=True, **kwargs)
        return lazy_func

    def draw_taghist(testres, **kwargs):
        e_ = 'taghist'
        func = find_expt_func(e_)
        ibs = testres.ibs
        build_commandline(e=e_, **kwargs)
        lazy_func = functools.partial(func, ibs, testres, **kwargs)
        return lazy_func

    def execute_test():
        func = find_expt_func(e)
        assert func is not None, 'unknown experiment e=%r' % (e,)

        argspec = ut.get_func_argspec(func)
        if (
            len(argspec.args) >= 2
            and argspec.args[0] == 'ibs'
            and argspec.args[1] == 'testres'
        ):
            # most experiments need a testres
            expts_kw = dict(
                defaultdb=db,
                dbdir=dbdir,
                a=a,
                t=t,
                qaid_override=qaid_override,
                daid_override=daid_override,
                initial_aids=initial_aids,
            )
            testdata_expts_func = functools.partial(
                main_helpers.testdata_expts, **expts_kw
            )

            ibs, testres = testdata_expts_func()
            # Build the requested drawing funciton
            draw_func = functools.partial(func, ibs, testres, **kwargs)
            testres.draw_func = draw_func
            ut.inject_func_as_method(testres, draw_cases)
            ut.inject_func_as_method(testres, draw_taghist)
            # testres.draw_cases = draw_cases
            return testres
        else:
            raise AssertionError('Unknown type of function for experiment')

    if lazy:
        return execute_test
    else:
        testres = execute_test()
        return testres


def testdata_expts(*args, **kwargs):
    ibs, testres = main_helpers.testdata_expts(*args, **kwargs)
    return testres


# import_subs()
# from wbia import gui
# from wbia import algo
# from wbia import templates
# from wbia import viz
# from wbia import web


# class _VizProxy(object):
#    def __init__(self):
#        pass

#    def getattr(self, key):
#        import wbia.viz as viz
#        return getattr(viz, key)

#    def setattr(self, key, val):
#        import wbia.viz as viz
#        return getattr(viz, key, val)


# viz = _VizProxy
# import apipkg
# apipkg.initpkg(__name__, {
#    'viz': {
#        'clone': "wbia.viz",
#    }
# }
# )

import logging
from wbia.init import main_helpers

testdata_cm = main_helpers.testdata_cm
testdata_cmlist = main_helpers.testdata_cmlist
testdata_qreq_ = main_helpers.testdata_qreq_
testdata_pipecfg = main_helpers.testdata_pipecfg
testdata_filtcfg = main_helpers.testdata_filtcfg
testdata_expts = main_helpers.testdata_expts
testdata_expanded_aids = main_helpers.testdata_expanded_aids
testdata_aids = main_helpers.testdata_aids

# Utool generated init makeinit.py
print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def reload_subs(verbose=True):
    """ Reloads wbia and submodules """
    import_subs()
    rrr(verbose=verbose)
    getattr(constants, 'rrr', lambda verbose: None)(verbose=verbose)
    getattr(entry_points, 'rrr', lambda verbose: None)(verbose=verbose)
    getattr(params, 'rrr', lambda verbose: None)(verbose=verbose)
    getattr(other, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(dbio, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(algo, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(control, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(viz, 'reload_subs', lambda: None)()

    getattr(gui, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(algo, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(viz, 'reload_subs', lambda verbose: None)(verbose=verbose)
    getattr(web, 'reload_subs', lambda verbose: None)(verbose=verbose)

    rrr(verbose=verbose)


rrrr = reload_subs


"""
Regen Command:
    Kinda have to work with the output of these. This module is hard to
    autogenerate correctly.

    cd /home/joncrall/code/wbia/wbia/other
    makeinit.py -x web viz tests gui
    makeinit.py -x constants params entry_points other control dbio tests
"""



# -*- coding: utf-8 -*-
import logging
import utool as ut
import sys

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def hack_argv(arg):
    # HACK COMMON ARGV SYMBOLS
    if arg.startswith('--hargv='):
        hack_argv_key = '='.join(arg.split('=')[1:])

        common_args = [
            '--dpath=~/latex/crall-thesis-2017/',
            '--clipwhite',
            '--contextadjust',
            '--dpi=256',
        ]

        sys.argv.extend(common_args)

        if '--noshow' not in sys.argv:
            sys.argv.extend(['--diskshow'])

        # Figsize hacks

        if hack_argv_key in ['surf']:
            sys.argv.extend(
                [
                    '--figsize=14,3',
                    '--hspace=.3',
                    # '--top=.8',
                    '--top=.85',
                    '--bottom=0.18',
                    '--left=.05',
                    '--right=.95',
                ]
            )
        elif hack_argv_key in ['scores']:
            sys.argv.extend(
                [
                    # '--figsize=15,7',
                    # '--top=.8',
                    '--figsize=14,3',
                    '--top=.8',
                    '--hspace=.3',
                    '--bottom=0.08',
                    '--left=.05',
                    '--right=.95',
                ]
            )
        elif hack_argv_key in ['tags']:
            sys.argv.extend(
                [
                    # '--figsize=14,3',
                    # '--top=.8',
                    # '--hspace=.3',
                    # '--bottom=0.08',
                    # '--left=.05',
                    # '--right=.95'
                ]
            )
        elif hack_argv_key in ['expt']:
            sys.argv.extend(
                [
                    # '--figsize=15,3',
                    '--figsize=15,3.1',
                    '--top=.9',
                    '--bottom=.15',
                ]
            )
        elif hack_argv_key in ['mech']:
            sys.argv.extend(
                [
                    # '--figsize=14,5',
                    '--figsize=14,3',
                    '--top=.9',
                ]
            )
        elif hack_argv_key in ['tags']:
            sys.argv.extend(['--bottom=.3', '--left=.2'])

        # Save location
        # fname_fmt = 'figuresX/expt_{e}_{db}_a_{a}_t_{t}'
        fname_fmt = 'figuresX/expt_{label}'

        # if hack_argv_key in ['scores']:
        #    fname_fmt += '_{filt}'

        # if hack_argv_key in ['time']:
        #    if not ('--falsepos' in sys.argv) or ('--truepos' in sys.argv):
        #        fname_fmt += '_TP'
        #    if ('--falsepos' in sys.argv):
        #        fname_fmt += '_FP'

        if hack_argv_key in ['time', 'expt', 'mech', 'scores', 'surf', 'tags']:
            sys.argv.extend(['--save', fname_fmt + '.png'])

        if hack_argv_key in ['time']:
            sys.argv.extend(['--figsize=18,8', r'--width=".8\textwidth"'])


# import IPython
# IPython.embed()
for arg in sys.argv[:]:
    hack_argv(arg)

import utool as ut  # NOQA

# ut.show_if_requested()
# sys.exit(1)
from utool.util_six import get_funcname  # NOQA

# import functools

# A list of registered development test functions
DEVCMD_FUNCTIONS = []
DEVPRECMD_FUNCTIONS = []

# DEVCMD_FUNCTIONS2 = []


# def devcmd2(*args):
#    """ Decorator which registers a function as a developer command """
#    noargs = len(args) == 1 and not isinstance(args[0], str)
#    if noargs:
#        # Function is only argument
#        func = args[0]
#        func_aliases = []
#    else:
#        func_aliases = list(args)
#    def closure_devcmd2(func):
#        global DEVCMD_FUNCTIONS2
#        func_aliases.extend([get_funcname(func)])
#        DEVCMD_FUNCTIONS2.append((tuple(func_aliases), func))
#        def func_wrapper(*args_, **kwargs_):
#            #if ut.VERBOSE:
#            #if ut.QUIET:
#            logger.info('[DEVCMD2] ' + ut.func_str(func, args_, kwargs_))
#            return func(*args_, **kwargs_)
#        return func_wrapper
#    if noargs:
#        return closure_devcmd2(func)
#    return closure_devcmd2


def devcmd(*args):
    """ Decorator which registers a function as a developer command """
    noargs = len(args) == 1 and not isinstance(args[0], str)
    if noargs:
        # Function is only argument
        func = args[0]
        func_aliases = []
    else:
        func_aliases = list(args)

    def closure_devcmd(func):
        global DEVCMD_FUNCTIONS
        func_aliases.extend([get_funcname(func)])
        DEVCMD_FUNCTIONS.append((tuple(func_aliases), func))

        def func_wrapper(*args_, **kwargs_):
            # if ut.VERBOSE:
            # if ut.QUIET:
            logger.info('[DEVCMD] ' + ut.func_str(func, args_, kwargs_))
            return func(*args_, **kwargs_)

        return func_wrapper

    if noargs:
        return closure_devcmd(func)
    return closure_devcmd


def devprecmd(*args):
    """ Decorator which registers a function as a developer precommand """
    noargs = len(args) == 1 and not isinstance(args[0], str)
    if noargs:
        # Function is only argument
        func = args[0]
        func_aliases = []
    else:
        func_aliases = list(args)

    def closure_devprecmd(func):
        global DEVPRECMD_FUNCTIONS
        func_aliases.extend([get_funcname(func)])
        DEVPRECMD_FUNCTIONS.append((tuple(func_aliases), func))

        def func_wrapper(*args_, **kwargs_):
            # if ut.VERBOSE:
            # if ut.QUIET:
            logger.info('[DEVPRECMD] ' + ut.func_str(func, args_, kwargs_))
            return func(*args_, **kwargs_)

        return func_wrapper

    if noargs:
        return closure_devprecmd(func)
    return closure_devprecmd



# -*- coding: utf-8 -*-
import logging
from os.path import join  # NOQA
import cv2
import numpy as np
import torch
import torch.nn
import utool as ut
import torchvision

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class LRSchedule(object):
    @staticmethod
    def exp(optimizer, epoch, init_lr=0.001, lr_decay_epoch=2):
        """Decay learning rate by a factor of 0.1 every lr_decay_epoch epochs."""
        lr = init_lr
        # epoch += 1
        if epoch % lr_decay_epoch == 0 and epoch != 0:
            lr *= 0.1

        if epoch % lr_decay_epoch == 0:
            logger.info('LR is set to {}'.format(lr))

        for param_group in optimizer.param_groups:
            param_group['lr'] = lr

        return optimizer, lr


def siam_vsone_train():
    r"""
    CommandLine:
        python -m wbia.algo.verif.torch.train_main siam_vsone_train

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.verif.torch.train_main import *  # NOQA
        >>> siam_vsone_train()
    """
    # wrapper around the RF vsone problem
    from wbia.algo.verif import vsone

    # pblm = vsone.OneVsOneProblem.from_empty('PZ_MTEST')
    pblm = vsone.OneVsOneProblem.from_empty('GZ_Master1')
    ibs = pblm.infr.ibs
    pblm.load_samples()
    samples = pblm.samples
    samples.print_info()
    xval_kw = pblm.xval_kw.asdict()
    skf_list = pblm.samples.stratified_kfold_indices(**xval_kw)

    def load_dataset(subset_idx):
        aids1, aids2 = pblm.samples.aid_pairs[subset_idx].T
        labels = pblm.samples['match_state'].y_enc[subset_idx]

        # train only on positive-vs-negative (ignore incomparable)
        labels = (labels == 1).astype(np.int64)

        chip_config = {'resize_dim': 'wh', 'dim_size': (224, 224)}
        img1_fpaths = ibs.depc_annot.get(
            'chips', aids1, read_extern=False, colnames='img', config=chip_config
        )
        img2_fpaths = ibs.depc_annot.get(
            'chips', aids2, read_extern=False, colnames='img', config=chip_config
        )
        dataset = LabeledPairDataset(img1_fpaths, img2_fpaths, labels)
        return dataset

    learn_idx, test_idx = skf_list[0]
    train_idx, val_idx = pblm.samples.subsplit_indices(learn_idx, n_splits=10)[0]

    # Split everything in the learning set into training / validation
    train_dataset = load_dataset(train_idx)
    vali_dataset = load_dataset(val_idx)
    test_dataset = load_dataset(test_idx)

    logger.info('* len(train_dataset) = {}'.format(len(train_dataset)))
    logger.info('* len(vali_dataset) = {}'.format(len(vali_dataset)))
    logger.info('* len(test_dataset) = {}'.format(len(test_dataset)))

    from wbia.algo.verif.torch import gpu_util

    gpu_num = gpu_util.find_unused_gpu(min_memory=6000)

    use_cuda = gpu_num is not None
    data_kw = {}
    if use_cuda:
        data_kw = {'num_workers': 6, 'pin_memory': True}
    batch_size = 64
    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True, **data_kw
    )
    vali_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True, **data_kw
    )
    test_loader = torch.utils.data.DataLoader(
        test_dataset, batch_size=batch_size, shuffle=False, **data_kw
    )

    from wbia.algo.verif.torch import fit_harness
    from wbia.algo.verif.torch import models
    from wbia.algo.verif.torch import netmath
    from wbia.algo.verif.torch import lr_schedule

    model = models.Siamese()

    criterion = netmath.Criterions.ContrastiveLoss(margin=1)
    lr_scheduler = lr_schedule.Exponential()
    optimizer_cls = netmath.Optimizers.Adam

    class_weights = train_dataset.class_weights()
    logger.info('class_weights = {!r}'.format(class_weights))

    harn = fit_harness.FitHarness(
        model=model,
        criterion=criterion,
        lr_scheduler=lr_scheduler,
        train_loader=train_loader,
        vali_loader=vali_loader,
        test_loader=test_loader,
        optimizer_cls=optimizer_cls,
        class_weights=class_weights,
        gpu_num=gpu_num,
    )
    harn.run()


class LabeledPairDataset(torch.utils.data.Dataset):
    """
    transform=transforms.Compose([
                       transforms.Scale(224),
                       transforms.ToTensor(),
                       torchvision.transforms.Normalize([0.5, 0.5, 0.5], [0.225, 0.225, 0.225])
                   ]

    Ignore:
        >>> from wbia.algo.verif.torch.train_main import *
        >>> from wbia.algo.verif.vsone import *  # NOQA
        >>> pblm = OneVsOneProblem.from_empty('PZ_MTEST')
        >>> ibs = pblm.infr.ibs
        >>> pblm.load_samples()
        >>> samples = pblm.samples
        >>> samples.print_info()
        >>> xval_kw = pblm.xval_kw.asdict()
        >>> skf_list = pblm.samples.stratified_kfold_indices(**xval_kw)
        >>> train_idx, test_idx = skf_list[0]
        >>> aids1, aids2 = pblm.samples.aid_pairs[train_idx].T
        >>> labels = pblm.samples['match_state'].y_enc[train_idx]
        >>> labels = (labels == 1).astype(np.int64)
        >>> chip_config = {'resize_dim': 'wh', 'dim_size': (224, 224)}
        >>> img1_fpaths = ibs.depc_annot.get('chips', aids1, read_extern=False, colnames='img', config=chip_config)
        >>> img2_fpaths = ibs.depc_annot.get('chips', aids2, read_extern=False, colnames='img', config=chip_config)
        >>> self = LabeledPairDataset(img1_fpaths, img2_fpaths, labels)
        >>> img1, img2, label = self[0]
    """

    def __init__(self, img1_fpaths, img2_fpaths, labels, transform='default'):
        assert len(img1_fpaths) == len(img2_fpaths)
        assert len(labels) == len(img2_fpaths)
        self.img1_fpaths = img1_fpaths
        self.img2_fpaths = img2_fpaths
        self.labels = labels
        if transform == 'default':
            transform = torchvision.transforms.Compose(
                [
                    # torchvision.transforms.Scale(224),
                    torchvision.transforms.ToTensor(),
                    torchvision.transforms.Normalize(
                        [0.5, 0.5, 0.5], [0.225, 0.225, 0.225]
                    ),
                ]
            )
        self.transform = transform

    def class_weights(self):
        import pandas as pd

        label_freq = pd.value_counts(self.labels)
        class_weights = label_freq.median() / label_freq
        class_weights = class_weights.sort_index().values
        class_weights = torch.from_numpy(class_weights.astype(np.float32))
        return class_weights

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image1, image2, label)
        """
        fpath1 = self.img1_fpaths[index]
        fpath2 = self.img2_fpaths[index]
        label = self.labels[index]

        def loader(fpath):
            bgr_255 = cv2.imread(fpath)
            bgr_01 = bgr_255.astype(np.float32) / 255.0
            rgb_01 = cv2.cvtColor(bgr_01, cv2.COLOR_BGR2RGB)
            return rgb_01

        img1 = loader(fpath1)
        img2 = loader(fpath2)

        if self.transform:
            img1 = self.transform(img1)
            img2 = self.transform(img2)

        return img1, img2, label

    def __len__(self):
        return len(self.img1_fpaths)
